#ifndef COMPARISON_H
#define COMPARISON_H

#include "DEFS.h"
#include "Schema.h"


// This structure stores an individual comparison that is part of a CNF
// Although defined as a class, its main role is grouping
// This is the reason that all the members are public
class Comparison {
public:
	//source of first comparison operand
	//target is the record and it can be Left, Right, or special record
	//whichAtt specifies the position of the comparison attribute in record schema
	Target operand1;
	int whichAtt1;

	//source of second comparison operand
	//similar to first compariosn operand
	Target operand2;
	int whichAtt2;

	//type of comparison operands
	Type attType;

	//comparison operator
	CompOperator op;

	Comparison();

	//copy constructor
	//copies the comparison structure from copy_me into the current object
	Comparison(const Comparison &copy_me);

	// print to the screen
	void Print ();
};



// This structure encapsulates a sort order for records
// A sort order consists of an ordered set of attributes from the record schema
// Records are sorted based on the order of the attributes
// Although defined as a class, its main role is grouping
// This is the reason that all the members are public
class OrderMaker {
public:
	//number of attributes from the schema used for sorting
	int numAtts;

	//positions and types of the attributes used for sorting
	//this specifies the ordered set of attributes used for sorting the records
	int whichAtts[MAX_ANDS];
	Type whichTypes[MAX_ANDS];

	
	OrderMaker();

	//the sorting order incorporates all the attributes from the record schema
	//Integer attributes are first, followed by Double attributes, and, finally,
	//String attributes
	OrderMaker(Schema *schema);

	//the sorting order is explicitly specified
	//join_atts contains the indices of the attributes appearing in the sort order
	//join_atts_no gives the number of attributes from the sort order
	//the sorting attributes are extracted from the schema
	OrderMaker(Schema *schema, int *join_atts, int join_atts_no);

	//copy constructor
	//the content of om1 is replicated in the current object
	OrderMaker(const OrderMaker &om1);

	//the union of the sort orders om1 and om2 is generated in the current object
	//it is assumed that om1 and om2 are built from the same schema
	//it is a set union because attributes are not duplicated
	void ANDMerge(OrderMaker *om1, OrderMaker *om2);


	// print to the screen
	void Print ();
};



// This structure stores a CNF expression that is to be evaluated during query execution
// Although defined as a class, its main role is grouping
// This is the reason that all the members are public
// A CNF is a comparison (logical) expression in conjunctive normal form
// It consists of (possibly) multiple disjunctive logical expressions connected with ANDs
// Multiple simple comparisons connected through ORs form a disjunctive logical formula
class CNF {
public:
	//bi-dimensional table of Comparison expressions
	//each row is a disjunctive formula consisting of multiple Comparisons connected
	//through ORs
	//all the rows are connected through ANDs
	Comparison orList[MAX_ANDS][MAX_ORS];

	//the size of each disjunctive formula or the number of Comparisons it contains
	int orLens[MAX_ANDS];

	//number of disjunctive formulas in the conjunctive big formula
	int numAnds;


	// this returns an instance of the OrderMaker class that
	// allows the CNF to be implemented using a sort-based
	// algorithm such as a sort-merge join.  Returns a 0 if and
	// only if it is impossible to determine an acceptable ordering
	// for the given comparison
	int GetSortOrders (OrderMaker &left, OrderMaker &right);

	//builds a new CNF out of from_me; all equi-join conditions identified by GetSortOrders
	//are eliminated, only the more intricate conditions being kept; return the number of AND
	//conditions kept after elimination
	int EliminateEquiJoinCond(CNF *from_me);

	//adds the predicate attributes for reading in C-STORE (the output attributes are in)
	//at the same time changes the CNF to be consistent with the new schema
	void AddCNFAtts(int *read_atts, int &read_atts_no);

	// print the comparison structure to the screen
	void Print ();
};

#endif
