#include "Controller.h"
#include "Statistics.h"
#include "SamplingEstimate.h"

#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>



Controller::Controller(int join_levels_no, LevelwiseStep **join_level, InMemoryJoinWrapper **imj_join_level, LevelwiseStep *aggregate_level, double conf_level, int groups_no, int tuples_stable, int var_estimates_needed)
{
	this->join_level = join_level;
	this->join_levels_no = join_levels_no;
	this->imj_join_level = imj_join_level;

	this->aggregate_level = aggregate_level;

	tuples_no = new (std::nothrow) double*[join_levels_no];
	if (tuples_no == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->conf_level = conf_level;

	this->groups_no = groups_no;

	if (this->groups_no == -1)
	{
		//in this case we have SPJ, no grouping
		stat_module = new (std::nothrow) Estimation_SPJ(tuples_stable, var_estimates_needed);
		if (stat_module == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	}
	else
	{
		//in this case we have grouping
		Schema *schema = this->aggregate_level->GetSchema(0);

		int *group_atts;
		int group_atts_no;
		this->aggregate_level->GetGroupAtts(0, group_atts, group_atts_no);

		group_schema = new (std::nothrow) Schema();
		if (group_schema == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		group_schema->CopyFrom(schema);
		group_schema->Project(group_atts, group_atts_no);

		stat_module = new (std::nothrow) Estimation_Group(tuples_stable, var_estimates_needed, this->join_levels_no, this->groups_no, group_schema);
		if (stat_module == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	}
}


Controller::~Controller()
{
	join_level = NULL;
	imj_join_level = NULL;

	aggregate_level = NULL;

	for (int i = 0; i < join_levels_no; i++)
		delete [] tuples_no[i];

	delete [] tuples_no;

	delete stat_module;
	stat_module = NULL;

	delete group_schema;
	group_schema = NULL;

	join_levels_no = 0;
}



void Controller::Control()
{
	gettimeofday(&stat_start_time, NULL);
	gettimeofday(&stat_last_time, NULL);

	Control_StartUp();

	Control_JoinLevel();

	Control_UpJoinLevel();

	if (groups_no == -1)
		Control_AggregateLevel();
	else
		Control_GroupAggregateLevel();


/*
	for (int i = 0; i < join_levels_no; i++)
	{
		for (int j = 0; j < join_level[i]->GetSize(); j++)
		{
			printf("Level %3d, Join %3d : %20.2f tuples\n", i, j, tuples_no[i][j]);
		}
	}
*/
}



void Controller::Control_StartUp()
{
	for (int i = 0; i < join_levels_no; i++)
		join_level[i]->StartUp();

	aggregate_level->StartUp();
}



void Controller::Control_JoinLevel()
{
	int i, j, k, n;

	int output_size;

	double *inP = new (std::nothrow) double[join_levels_no - 1];
	if (inP == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	FILE *file_p_in = fopen("INP.DAT", "rt");
	for (i = 0; i < join_levels_no - 1; i++)
	{
		char line[80];
		fscanf(file_p_in, "%s", line);
		inP[i] = atof(line);
	}
	fclose(file_p_in);
	

	for (j = 0; j < join_levels_no - 1; j++)
	{
		n = join_level[j]->GetSize();

		tuples_no[j] = new (std::nothrow) double[n];
		if (tuples_no[j] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		double *p = new (std::nothrow) double[n];
		if (p == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		double *cur_tot = new (std::nothrow) double[n];
		if (cur_tot == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (i = 0; i < n; i++)
		{
			tuples_no[j][i] = 0;
			p[i] = inP[j];
		}


		//add this level to the statistics module
		stat_module->AddLevel(n, 0);



		for (i = 0; i < n; i++)
		{
			int buffer_size = join_level[j]->out_pipe_sizes[i];

			Record *match_list = new (std::nothrow) Record[buffer_size];
			if (match_list == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			int ret = join_level[j]->GetNextBatch(i, p[i], match_list, buffer_size, output_size);
			tuples_no[j][i] += output_size;

			Record *out_imj_list = NULL;
			imj_join_level[j]->Swap(match_list, out_imj_list, i, output_size);

			delete [] out_imj_list;

			cur_tot[i] = join_level[j]->HowMuchIsLeft(i);
		}


		SearchResults *results = imj_join_level[j]->Join(p, groups_no);


		SimpleSamplingData *sampling_prop = NULL;
		if (j == 0)
		{
			int *tuples_now = new (std::nothrow) int[n];
			if (tuples_now == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			for (k = 0; k < n; k++)
				tuples_now[k] = (int)(p[k] * join_level[0]->relation_size[k]);

			sampling_prop = new (std::nothrow) SimpleSamplingData(n, join_level[j]->relation_size, tuples_now);
			if (sampling_prop == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			delete [] tuples_now;
		}
		else
		{
			sampling_prop = new (std::nothrow) SimpleSamplingData(n, p);
			if (sampling_prop == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
		}

		stat_module->ProcessTuples(results, sampling_prop);

		delete sampling_prop;
		delete results;



		// check to see if it is time to give out another result
		gettimeofday(&stat_current_time, NULL);
		unsigned int t_time = stat_current_time.tv_sec - stat_last_time.tv_sec;
		if (t_time >= UPDATE_FREQ)
		{
//			stat_module->Print_Level();

			// write the result to the file!
			stat_last_time = stat_current_time;

			int estimates_no;
			double *est, *var, *low, *high;
			Record *groups_id;

			stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level);

			t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

			for (i = 0; i < n; i++)
			{
				printf("Level %3d, Join %3d : left %20.18f !!!\n", j, i, cur_tot[i]);
			}

			if (groups_no == -1)
			{
				printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f\n", j, t_time, *low, *est, *high, *var);
			}
			else
			{
				for (i = 0; i < estimates_no; i++)
				{
					char *group_info;
					groups_id[i].Print(group_schema, group_info);
					printf("%s\n", group_info);

					printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f\n", j, t_time, low[i], est[i], high[i], var[i]);

					delete [] group_info;
				}
			}

			delete [] est;
			delete [] var;
			delete [] low;
			delete [] high;
			delete [] groups_id;
		}



		while (1)
		{
			int still_running = 0;

			for (i = 0; i < n; i++)
			{
				int buffer_size = join_level[j]->out_pipe_sizes[i];

				Record *match_list = new (std::nothrow) Record[buffer_size];
				if (match_list == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				double in_l = join_level[j]->HowMuchIsLeft(i);

				int ret = join_level[j]->GetNextBatch(i, p[i], match_list, buffer_size, output_size);
				tuples_no[j][i] += output_size;

				if (ret == 0)
				{
					delete [] match_list;
					continue;
				}
				else if (ret == 1)
				{
					still_running = 1;
				}

				Record *out_imj_list = NULL;
				imj_join_level[j]->Swap(match_list, out_imj_list, i, output_size);

				//deal separately with the last portion which could be smaller than expected
				cur_tot[i] = join_level[j]->HowMuchIsLeft(i);
				if (cur_tot[i] == 0.0)
				{
					p[i] = in_l;
				}


				SimpleSamplingData *sampling_prop = NULL;
				if (j == 0)
				{
					int *tuples_now = new (std::nothrow) int[n];
					if (tuples_now == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					for (k = 0; k < n; k++)
						tuples_now[k] = (int)(p[k] * join_level[0]->relation_size[k]);

					sampling_prop = new (std::nothrow) SimpleSamplingData(n, join_level[j]->relation_size, tuples_now);
					if (sampling_prop == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					delete [] tuples_now;
				}
				else
				{
					sampling_prop = new (std::nothrow) SimpleSamplingData(n, p);
					if (sampling_prop == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
				}

				SearchResults *results = imj_join_level[j]->Join(p, groups_no);

				stat_module->ProcessTuples(results, sampling_prop);

				delete sampling_prop;
				delete results;

	
				delete [] out_imj_list;
			}
	

			// check to see if it is time to give out another result
			gettimeofday(&stat_current_time, NULL);
			unsigned int t_time = stat_current_time.tv_sec - stat_last_time.tv_sec;
			if (t_time >= UPDATE_FREQ)
			{

//				stat_module->Print_Level();

				// write the result to the file!
				stat_last_time = stat_current_time;

				int estimates_no;
				double *est, *var, *low, *high;
				Record *groups_id;
	
				stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level);

				t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

				for (i = 0; i < n; i++)
				{
					printf("Level %3d, Join %3d : left %20.18f !!!\n", j, i, cur_tot[i]);
				}

				if (groups_no == -1)
				{
					printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f\n", j, t_time, *low, *est, *high, *var);
				}
				else
				{
					for (i = 0; i < estimates_no; i++)
					{
						char *group_info;
						groups_id[i].Print(group_schema, group_info);
						printf("%s\n", group_info);

						printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f\n", j, t_time, low[i], est[i], high[i], var[i]);

						delete [] group_info;
					}
				}
	
				delete [] est;
				delete [] var;
				delete [] low;
				delete [] high;
				delete [] groups_id;
			}


			if (still_running == 0)
				break;
		}


		delete [] p;
		delete [] cur_tot;

		for (int jj = 0; jj < join_level[j]->GetSize(); jj++)
		{
			printf("Level %3d, Join %3d : %20.2f tuples\n", j, jj, tuples_no[j][jj]);
		}

		delete join_level[j];
		join_level[j] = NULL;

		delete imj_join_level[j];
		imj_join_level[j] = NULL;
	}

	delete [] inP;
}


void Controller::Control_UpJoinLevel()
{
	int i, n;

	int output_size;

	n = join_level[join_levels_no - 1]->GetSize();

	double *p = new (std::nothrow) double[n];
	if (p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	tuples_no[join_levels_no - 1] = new (std::nothrow) double[n];
	if (tuples_no[join_levels_no - 1] == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *cur_tot = new (std::nothrow) double[n];
	if (cur_tot == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (i = 0; i < n; i++)
	{
		p[i] = -1.0;
		tuples_no[join_levels_no - 1][i] = 0;
	}



	//add this level to the statistics module
	stat_module->AddGeneralLevel(n, 1);

	double tot_tuples = 0.0;

	while (1)
	{
		int still_running = 0;


		for (i = 0; i < n; i++)
		{
			int buffer_size = join_level[join_levels_no - 1]->out_pipe_sizes[i];

			Record *match_list = new (std::nothrow) Record[buffer_size];
			if (match_list == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			double in_l = join_level[join_levels_no - 1]->HowMuchIsLeft(i);

			int ret = join_level[join_levels_no - 1]->GetNextBatch(i, p[i], match_list, buffer_size, output_size);
			tuples_no[join_levels_no - 1][i] += output_size;

			if (ret == 0)
			{
				delete [] match_list;
				continue;
			}
			else
			{
				still_running = 1;
			}


			//build the aggregate function values for the returned tuples
			double *aggregate = new (std::nothrow) double[output_size];
			if (aggregate == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			Function *func = aggregate_level->GetFunction(0);
			for (int j = 0; j < output_size; j++)
			{
				int intResult;
				double doubleResult;
		
				Type t = func->Apply(&match_list[j], intResult, doubleResult);
		
				if (t == Int)
					aggregate[j] = (double)intResult;
				else if (t == Double)
					aggregate[j] = doubleResult;
			}


//			cout << "Tuples returned: " << output_size << "\n";



			//collapsing goes in here
			Schema *agg_schema = aggregate_level->GetSchema(0);

			int *join_atts;
			int join_atts_no;
			aggregate_level->GetJoinAtts(0, join_atts, join_atts_no);

			OrderMaker *join_order = new (std::nothrow) OrderMaker(agg_schema, join_atts, join_atts_no);
			if (join_order == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			int tup_no;

			if (groups_no == -1)
			{
				tup_no = Collaps_Aggregate_Level(match_list, aggregate, output_size, join_order);
//				int tup_no = Collaps_Aggregate_Level(match_list, aggregate, output_size);
//				int tup_no = output_size;
				tot_tuples += tup_no;
	
				delete join_order;
			}
			else
			{
				//in this case we have grouping
				//the grouping attributes have to be added to the OrderMaker
				int *group_atts;
				int group_atts_no;
				aggregate_level->GetGroupAtts(0, group_atts, group_atts_no);
	
				OrderMaker *group_order = new (std::nothrow) OrderMaker(agg_schema, group_atts, group_atts_no);
				if (group_order == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				OrderMaker *mixed_order = new (std::nothrow) OrderMaker();
				if (mixed_order == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				mixed_order->ANDMerge(join_order, group_order);

				delete group_order;
				delete join_order;

				tup_no = Collaps_Group_Level(match_list, aggregate, output_size, mixed_order);
				tot_tuples += tup_no;

				for (int j = 0; j < tup_no; j++)
				{
					match_list[j].Project(group_atts, group_atts_no, agg_schema->GetNumAtts());
				}

				delete mixed_order;
			}


//			cout << "Tuples collapsed: " << tup_no << "\n";


			cur_tot[i] = 1.0 - join_level[join_levels_no - 1]->HowMuchIsLeft(i);
//			printf("UpLevel, Join %3d : left %20.18f !!!\n", i, 1.0 - cur_tot[i]);


//statistics stuff gets in here
			SimpleSamplingData *sampling_prop = new (std::nothrow) SimpleSamplingData(n, cur_tot);
			if (sampling_prop == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			if (groups_no == -1)
			{
				stat_module->ProcessTuples(aggregate, NULL, tup_no, sampling_prop);
			}
			else
			{
				stat_module->ProcessTuples(aggregate, match_list, tup_no, sampling_prop);
			}

			delete sampling_prop;
//statistics stuff end


/*
			//the new collapsing code goes in here
			//first build the collaps structure
			SearchResults *myResult = new (std::nothrow) SearchResults;
			if (myResult == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			myResult->numConst = 1;

			myResult->constRels = new int;
			*myResult->constRels = 0;

			myResult->ratios = new double;
			*myResult->ratios = 1.0;

			myResult->numDel = 0;
			myResult->funcDel = NULL;
			myResult->delTuples = NULL;

			myResult->numRes = output_size;
			myResult->funcRes = aggregate;

			myResult->resTuples = new (std::nothrow) int*[myResult->numRes];
			if (myResult->resTuples == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			for(int j = 0; j < myResult->numRes; j++)
			{
				myResult->resTuples[j] = new int;
				*myResult->resTuples[j] = match_list[j].hash;
			}


			cur_tot[i] = 1.0 - join_level[join_levels_no - 1]->HowMuchIsLeft(i);
//			printf("UpLevel, Join %3d : left %20.18f !!!\n", i, 1.0 - cur_tot[i]);

			stat_module->ProcessTuples(myResult, cur_tot[0]);

			delete myResult;
*/



			delete [] match_list;
			delete [] aggregate;
		}

		// check to see if it is time to give out another result
		gettimeofday(&stat_current_time, NULL);
		unsigned int t_time = stat_current_time.tv_sec - stat_last_time.tv_sec;
		if (t_time >= UPDATE_FREQ)
		{
			// write the result to the file!
			stat_last_time = stat_current_time;

			int estimates_no;
			double *est, *var, *low, *high;
			Record *groups_id;
	
			stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level);

			t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

			if (groups_no == -1)
			{
				printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, *low, *est, *high, *var, cur_tot[0]);
			}
			else
			{
				for (i = 0; i < estimates_no; i++)
				{
					char *group_info;
					groups_id[i].Print(group_schema, group_info);
					printf("%s\n", group_info);

					printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, low[i], est[i], high[i], var[i], cur_tot[0]);

					delete [] group_info;
				}
			}
	
			delete [] est;
			delete [] var;
			delete [] low;
			delete [] high;
			delete [] groups_id;
		}

		if (still_running == 0)
			break;
	}

	delete [] p;
	delete [] cur_tot;


//last call for estimates before exiting
	int estimates_no;
	double *est, *var, *low, *high;
	Record *groups_id;

	stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level);

	gettimeofday(&stat_current_time, NULL);
	unsigned int t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

	cout << "End Level: " << join_levels_no - 1 << "\n";

	if (groups_no == -1)
	{
		printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, *low, *est, *high, *var, 1.0);
	}
	else
	{
		for (i = 0; i < estimates_no; i++)
		{
			char *group_info;
			groups_id[i].Print(group_schema, group_info);
			printf("%s\n", group_info);

			printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, low[i], est[i], high[i], var[i], 1.0);

			delete [] group_info;
		}
	}

	delete [] est;
	delete [] var;
	delete [] low;
	delete [] high;
	delete [] groups_id;


	for (int jj = 0; jj < join_level[join_levels_no - 1]->GetSize(); jj++)
	{
		printf("Level %3d, Join %3d : %20.2f tuples\n", join_levels_no - 1, jj, tuples_no[join_levels_no - 1][jj]);
	}


//	cout << "Total Tuples : " << tot_tuples << "\n";


	delete join_level[join_levels_no - 1];
	join_level[join_levels_no - 1] = NULL;
}


void Controller::Control_AggregateLevel()
{
	int i, n;

	n = aggregate_level->GetSize();
	for (i = 0; i < n; i++)
	{
		//fake call to GetNextBatch for synchronization purposes
		int output_size;
		int ret = aggregate_level->GetNextBatch(i, 0.0, NULL, -1, output_size);

		double aggregate = aggregate_level->GetAggregate(i);
		double tup_no = aggregate_level->GetTuplesNo(i);

		printf("Result:%30.4f\n", aggregate);
		printf("Tuples:%30.1f\n\n", tup_no);
	}
}


void Controller::Control_GroupAggregateLevel()
{
	int i, n;

	n = aggregate_level->GetSize();
	for (i = 0; i < n; i++)
	{
		//fake call to GetNextBatch for synchronization purposes
		int output_size;
		int ret = aggregate_level->GetNextBatch(i, 0.0, NULL, -1, output_size);

		Schema *schema = aggregate_level->GetSchema(i);

		int *group_atts;
		int group_atts_no;
		aggregate_level->GetGroupAtts(i, group_atts, group_atts_no);

		schema->Project(group_atts, group_atts_no);

		vector<Record*> desc;
		vector<double> aggregate;
		vector<double> tuples;

		int extract_groups_no;

		aggregate_level->GetGroups(i, desc, aggregate, tuples, extract_groups_no);

		for (int j = 0; j < extract_groups_no; j++)
		{
			desc[j]->Print(schema);
			printf("Result: %30.4f\n", aggregate[j]);
			printf("Tuples: %30.1f\n", tuples[j]);
		}

		printf("Groups No: %30.1f\n\n", (double)extract_groups_no);
	}
}





//collapsing method for the aggregate level; it aggregates the function values for each of the distinct groups
int Collaps_Aggregate_Level(Record *join_res, double *&output, int v_size, OrderMaker *join_order)
//int Collaps_Aggregate_Level(Record *join_res, double *&output, int v_size)
{
	Hash *eq_func = new (std::nothrow) Hash();
	if (eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *new_output = new (std::nothrow) double[v_size];
	if (new_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	vector<int> index;
	vector<double> sum_func;

	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		index.push_back(c_i);
		sum_func.push_back(output[c_i]);

//		printf("HASH : %30d\n", join_res[c_i].hash);

		c_j = c_i + 1;
		while ((c_j < v_size) && (join_res[c_i].hash == join_res[c_j].hash))
		{
			int is_in = 0;
			int nn = index.size();

			for (int i = 0; i < nn; i++)
			{
				if (eq_func->Compare(&join_res[index[i]], &join_res[c_j], join_order) == 0)
				{
					//check to see if this record already exists
					sum_func[i] = sum_func[i] + output[c_j];

					is_in = 1;
					break;
				}
			}

			if (is_in == 0)
			{
				//in this case there is a conflict in the hash
				index.push_back(c_j);
				sum_func.push_back(output[c_j]);

//				cout << "Conflict!!!\n";
			}

			c_j += 1;
		}

		int nn = index.size();
		for (int i = 0; i < nn; i++)
		{
			new_output[colaps_count + i] = sum_func[i];
		}

		colaps_count += nn;

		index.clear();
		sum_func.clear();

		c_i = c_j;
	}


	delete [] output;
	output = new_output;

	delete eq_func;

	return colaps_count;



/*
	Hash *eq_func = new (std::nothrow) Hash();
	if (eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *new_output = new (std::nothrow) double[v_size];
	if (new_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		new_output[colaps_count] = output[c_i];

		c_j = c_i + 1;


		while ((c_j < v_size) && (eq_func->Compare(&join_res[c_i], &join_res[c_j], join_order) == 0))
		{
			new_output[colaps_count] += output[c_j];

			c_j += 1;
		}

		c_i = c_j;
		colaps_count += 1;
	}


	delete [] output;
	output = new_output;

	delete eq_func;

	return colaps_count;
*/

}



//collapsing method for the aggregate level; it aggregates the function values for each of the distinct groups
int Collaps_Group_Level(Record *&join_res, double *&output, int v_size, OrderMaker *join_order)
{
	Hash *eq_func = new (std::nothrow) Hash();
	if (eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *new_output = new (std::nothrow) double[v_size];
	if (new_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Record *new_join_res = new (std::nothrow) Record[v_size];
	if (new_join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	vector<int> index;
	vector<double> sum_func;

	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		index.push_back(c_i);
		sum_func.push_back(output[c_i]);

//		printf("HASH : %30d\n", join_res[c_i].hash);

		c_j = c_i + 1;
		while ((c_j < v_size) && (join_res[c_i].hash == join_res[c_j].hash))
		{
			int is_in = 0;
			int nn = index.size();

			for (int i = 0; i < nn; i++)
			{
				if (eq_func->Compare(&join_res[index[i]], &join_res[c_j], join_order) == 0)
				{
					//check to see if this record already exists
					sum_func[i] = sum_func[i] + output[c_j];

					is_in = 1;
					break;
				}
			}

			if (is_in == 0)
			{
				//in this case there is a conflict in the hash
				index.push_back(c_j);
				sum_func.push_back(output[c_j]);

//				cout << "Conflict!!!\n";
			}

			c_j += 1;
		}

		int nn = index.size();
		for (int i = 0; i < nn; i++)
		{
			new_output[colaps_count + i] = sum_func[i];
			new_join_res[colaps_count + i].Consume(&join_res[index[i]]);
		}

		colaps_count += nn;

		index.clear();
		sum_func.clear();

		c_i = c_j;
	}


	delete [] output;
	output = new_output;

	delete [] join_res;
	join_res = new_join_res;

	delete eq_func;

	return colaps_count;
}




//collapsing method based on linear hashing; it aggregates the function values for each of the distinct groups
int Collaps_Hash(int **&join_res, double *&output, int v_size, int relations_no)
{

	//THIS IMPLEMENTATION IS BASED ON HASHING//

	int *colaps_array = new (std::nothrow) int[v_size];
	if (colaps_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < v_size; i++)
	{
		colaps_array[i] = -1;
	}

	int colaps_count = 0;
	for (int i = 0; i < v_size; i++)
	{
		//compute the index in the hash array of each record in the output

		unsigned int index = 0;
		for (int j = 0; j < relations_no; j++)
		{
			index += join_res[i][j];

		}
		index = index % v_size;

		if (colaps_array[index] == -1)
		{
			//in this case the position is free, so just add the index to the array
			colaps_array[index] = i;
			colaps_count += 1;

		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = memcmp(join_res[i], join_res[colaps_array[index]], relations_no * sizeof(int));
			if (are_identical == 0)
			{
				//the records are identical, so collapse the aggregate values
				output[colaps_array[index]] += output[i];

			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % v_size; (colaps_array[j] != -1) && (j != index); j = (j + 1) % v_size)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = memcmp(join_res[i], join_res[colaps_array[j]], relations_no * sizeof(int));
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					output[colaps_array[j]] += output[i];
				}
				else
				{
					if (colaps_array[j] == -1)
					{
						colaps_array[j] = i;
						colaps_count += 1;
					}
					else
					{
						cout << "This should be an error. No place was found in the hash array for record with index " << i << " in the result.";
						exit(1);
					}
				}
			}
		}
	}

	int **p_join_res = new (std::nothrow) int*[colaps_count];
	if (p_join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *p_output = new (std::nothrow) double[colaps_count];
	if (p_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

//	double tot_res = 0.0;

	int c_i = 0;
	for (int i = 0; i < v_size; i++)
	{
		if (colaps_array[i] != -1)
		{
			p_join_res[c_i] = new (std::nothrow) int[relations_no];
			if (p_join_res[c_i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			memcpy(p_join_res[c_i], join_res[colaps_array[i]], relations_no * sizeof(int));
	
			p_output[c_i] = output[colaps_array[i]];

//			tot_res += p_output[c_i];

			c_i += 1;
			if (c_i > colaps_count)
			{
				cout << "Error with putting the collapsed records together !";
				exit(1);
			}
		}

		delete [] join_res[i];
	}

	delete [] join_res;
	join_res = p_join_res;

	delete [] output;
	output = p_output;


	delete [] colaps_array;

	//end of collapsing


// 	printf("InMemoryJoinResult: %30.5f\n\n", tot_res);


	return colaps_count;
}



//collapsing method based on sorting; it aggregates the function values for each of the distinct groups
int Collaps_Sort(int **&join_res, double *&output, int v_size, int relations_no)
{
	//THIS IMPLEMENTATION IS BASED ON SORTING//

	int *colaps_array = new (std::nothrow) int[v_size * (relations_no + 2)];
	if (colaps_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < v_size; i++)
	{
		colaps_array[i * (relations_no + 2)] = relations_no;

		memcpy(colaps_array + i * (relations_no + 2) + 1, join_res[i], relations_no * sizeof(int));

		colaps_array[i * (relations_no + 2) + relations_no + 1] = i;

		delete [] join_res[i];
	}

	delete [] join_res;

	qsort(colaps_array, v_size, sizeof(int) * (relations_no + 2), Compare_Direct_IMJ);

	int *colaps_index = new (std::nothrow) int[v_size];
	if (colaps_index == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		colaps_index[colaps_count] = c_i;
		colaps_count += 1;

		c_j = c_i + 1;


		while ((c_j < v_size) && (memcmp(colaps_array + c_i * (relations_no + 2), colaps_array + c_j * (relations_no + 2), (relations_no + 1) * sizeof(int)) == 0))
		{
			output[colaps_array[c_i * (relations_no + 2) + relations_no + 1]] += output[colaps_array[c_j * (relations_no + 2) + relations_no + 1]];

			c_j += 1;
		}

		c_i = c_j;
	}


	//here we have in colaps_index colaps_count records that have survived

	join_res = new (std::nothrow) int*[colaps_count];
	if (join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *p_output = new (std::nothrow) double[colaps_count];
	if (p_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


//	double tot_res = 0.0;


	for (int i = 0; i < colaps_count; i++)
	{
		join_res[i] = new (std::nothrow) int[relations_no];
		if (join_res[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		memcpy(join_res[i], colaps_array + colaps_index[i] * (relations_no + 2) + 1, relations_no * sizeof(int));

		p_output[i] = output[colaps_array[colaps_index[i] * (relations_no + 2) + relations_no + 1]];

//		tot_res += p_output[i];
	}
	//done with putting the results in the output


	delete [] colaps_index;

	delete [] colaps_array;

	delete [] output;
	output = p_output;

	//end of collapsing

// 	printf("InMemoryJoinResult: %30.5f\n\n", tot_res);


	return colaps_count;
}



int Compare_IMJ(const void *r1, const void *r2)
{
	int *a, *b;

	a = (int*)*(int*)r1;
	b = (int*)*(int*)r2;

	for (int i = 1; i <= *a; i++)
	{
		if (*(a + i) > *(b + i))
			return 1;
		else if (*(a + i) < *(b + i))
			return -1;
	}

	return 0;
}


int Compare_Direct_IMJ(const void *r1, const void *r2)
{
	int *a, *b;

	a = (int*)r1;
	b = (int*)r2;

	int res = memcmp(a + 1, b + 1, *a * sizeof(int));
	return res;

/*
	for (int i = 1; i <= *a; i++)
	{
		if (*(a + i) > *(b + i))
			return 1;
		else if (*(a + i) < *(b + i))
			return -1;
	}

	return 0;
*/

}



