#include "ControllerTurbo.h"
#include "Statistics.h"
#include "SamplingEstimate.h"
#include "ControllerIQ.h"
#include "RecPriorityQ.h"
#include "DEFS.h"

#include <iostream>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>


pthread_mutex_t xyzmutex;
void handler (int signum) {
	pthread_mutex_unlock (&xyzmutex);
}



Waiter::Waiter () {
	pthread_mutex_init (&xyzmutex, NULL);
	pthread_mutex_lock (&xyzmutex);
	signal (SIGALRM, handler);
}


void Waiter::Wait (int msec) {
	myInterval.it_value.tv_sec = msec / 1000;
	myInterval.it_value.tv_usec = 1000 * (msec % 1000);
	myInterval.it_interval.tv_sec = myInterval.it_interval.tv_usec = 0;

	setitimer (ITIMER_REAL, &myInterval, 0);
	pthread_mutex_lock (&xyzmutex);
}




ControllerTurbo::ControllerTurbo(int join_levels_no, LevelwiseStep **join_level, NewInMemoryJoin **imj_join_level, LevelwiseStep *aggregate_level, double conf_level, int groups_no, int tuples_stable, int var_estimates_needed, int queue_size)
{
	this->join_level = join_level;
	this->join_levels_no = join_levels_no;
	this->imj_join_level = imj_join_level;

	this->aggregate_level = aggregate_level;

	tuples_no = new (std::nothrow) double*[join_levels_no];
	if (tuples_no == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->conf_level = conf_level;

	this->groups_no = groups_no;

	if (this->groups_no == -1)
	{
		//in this case we have SPJ, no grouping
		stat_module = new (std::nothrow) Estimation_SPJ(tuples_stable, var_estimates_needed);
		if (stat_module == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	}
	else
	{
		//in this case we have grouping
		Schema *schema = this->aggregate_level->GetSchema(0);

		int *group_atts;
		int group_atts_no;
		this->aggregate_level->GetGroupAtts(0, group_atts, group_atts_no);

		group_schema = new (std::nothrow) Schema();
		if (group_schema == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		group_schema->CopyFrom(schema);
		group_schema->Project(group_atts, group_atts_no);

		stat_module = new (std::nothrow) Estimation_Group(tuples_stable, var_estimates_needed, this->join_levels_no, this->groups_no, group_schema);
		if (stat_module == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	}

	this->queue_size = queue_size;
}


ControllerTurbo::~ControllerTurbo()
{
	join_level = NULL;
	imj_join_level = NULL;

	aggregate_level = NULL;

	for (int i = 0; i < join_levels_no; i++)
		delete [] tuples_no[i];

	delete [] tuples_no;

	delete stat_module;
	stat_module = NULL;

	delete group_schema;
	group_schema = NULL;

	join_levels_no = 0;
}



void ControllerTurbo::Control()
{
	gettimeofday(&stat_start_time, NULL);
	gettimeofday(&stat_last_time, NULL);

	Control_StartUp();

	Control_JoinLevel();

	Control_UpJoinLevel();

	if (groups_no == -1)
		Control_AggregateLevel();
	else
		Control_GroupAggregateLevel();

/*
	for (int i = 0; i < join_levels_no; i++)
	{
		for (int j = 0; j < join_level[i]->GetSize(); j++)
		{
			printf("Level %3d, Join %3d : %20.2f tuples\n", i, j, tuples_no[i][j]);
		}
	}
*/
}



void ControllerTurbo::Control_StartUp()
{
	for (int i = 0; i < join_levels_no; i++)
		join_level[i]->StartUp();

	aggregate_level->StartUp();
}



void ControllerTurbo::Control_JoinLevel()
{
	int i, j, k, n;

	int output_size;

	for (j = 0; j < join_levels_no - 1; j++)
	{
		n = join_level[j]->GetSize();

		tuples_no[j] = new (std::nothrow) double[n];
		if (tuples_no[j] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		//initialize the level input queues
		ControllerInputQueue **i_queues = new (std::nothrow) ControllerInputQueue*[n];
		if (i_queues == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (i = 0; i < n; i++)
		{
			i_queues[i] = new (std::nothrow) ControllerInputQueue(join_level[j]->operators[i], join_level[j]->out_pipe_sizes[i], queue_size);
			if (i_queues[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			tuples_no[j][i] = 0.0;
		}

		//add this level to the statistics module
		stat_module->AddGeneralLevel(n, 0);


		//pass the statistics module to the IMJ for this level
		imj_join_level[j]->SetStatModule(stat_module);

		// this one is used internally
		double *samplingRatio = new double[n];
		double threshold;
		int timeStamp;

		SearchResults Result;
		Result.numRes = 0;
		Result.numDel = 0;
		



		// this loop will wait for TOTWAITTIME seconds, or until
		// all of the input pipes are done (whichever comes first).
		// Basically, it periodically wakes up to see if all pipes
		// have been exhausted.  If so, then it will exit early

		time_t lastTime = time (NULL);
		for (i = 0; 1; i++)
		{
			Waiter foo;
			foo.Wait (TOTWAITTIME * 10);
	
			// see if all of the pipes have shut down
			int allDone = 1;
			for (int j = 0; j < n; j++) {
				if (i_queues[j]->IsConstant() == 0) {
					allDone = 0;
					break;
				}
			}
	
			// if we are all done, then exit this loop early
			if (allDone == 1)
				break;

			time_t this_time = time(NULL);
			// in this case, we have met or exceeded our startup
			// time.  We must keep looping until each relation either
			// (a) has at least one tuple, or (b) is constant.  This
			// loop checks to see if this has happened
			if (this_time - lastTime > TOTWAITTIME) {
				int allConstantOrHaveTuples = 1;
				for (int j = 0; j < n; j++) {
					if (i_queues[j]->IsConstant() == 0) {
						if (i_queues[j]->GetNumTuples() == 0) {
							allConstantOrHaveTuples = 0;
							break;
                        }
                    }
	            }			

                if (allConstantOrHaveTuples)
                    break;
			}
		}


/* Now, we are done with the startup phase.  We can next determine the constant relations and send them to the IMJ: */
		// this one is used to send to the IMJ
		int *constants = new int[n];
	
		// this one is used internally
		int *isConstant = new int[n];
	
		int numDone = 0;
		for (i = 0; i < n; i++)
		{
			if (i_queues[i]->IsConstant() == 1) {
				constants[numDone] = i;
				numDone += 1;
				isConstant[i] = 1;

//				printf("Level %3d, Join %3d : CONSTANT\n", j, i);
			}
			else {
				isConstant[i] = 0;
			}
		}
	
		imj_join_level[j]->InitialConstRel(numDone, constants);
		delete [] constants;


/* Now we can send the constant tuples to the IMJ: */
		for (i = 0; i < n; i++)
		{
			// if this one is a constant, then load him into the IMJ
			if (isConstant[i])
			{
				// build the array of tuples
				int numTuples = i_queues[i]->GetNumTuples();
				if (numTuples <= 0)
				{
					printf("CONSTANT RELATION WITH 0 TUPLES !!! EXIT !!!\n");
					exit(0);
				}

				printf("Level %3d, Join %3d IS CONSTANT : %20d !!!\n", j, i, numTuples);

				Record *recs = new (std::nothrow) Record[numTuples];
				if (recs == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
		
				int ts;
				for (k = 0; k < numTuples; k++) {
					i_queues[i]->GetNextTuple(&recs[k], ts);
				}
	
				// add them into the IMJ
				imj_join_level[j]->AddConstTuples(recs, numTuples, i);

				tuples_no[j][i] += numTuples;
	
				// delete the memory
				delete [] recs;

				samplingRatio[i] = 1.0;
			}
	
		}



		if (numDone == n)
		{
			//this is a special case in which all the relations are constant
			// at this point, the IMJ is ready to go off and choose a permutation order

//			SearchResults *myResult = imj_join_level[j]->ChoosePermOrder(1.0);
			SearchResults *myResult = imj_join_level[j]->ChoosePermOrder();

			printf("The permutation order was determined. Starting the normal execution !!!\n");

			if (myResult != NULL)
			{
				stat_module->ProcessTuples(myResult, 1.0);

				Result.numRes += myResult->numRes;
				Result.numDel += myResult->numDel;


				delete myResult;
			}

			int estimates_no;
			double *est, *var, *low, *high;
			Record *groups_id;

			stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level, 1.0);


			gettimeofday(&stat_current_time, NULL);
			unsigned int t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

			if (groups_no == -1)
			{
				printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", j, t_time, *low, *est, *high, *var, 1.0);
			}
			else
			{
				for (i = 0; i < estimates_no; i++)
				{
					char *group_info;
					groups_id[i].Print(group_schema, group_info);
					printf("%s\n", group_info);

					printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", j, t_time, low[i], est[i], high[i], var[i], 1.0);

					delete [] group_info;
				}
			}

			delete [] est;
			delete [] var;
			delete [] low;
			delete [] high;
			delete [] groups_id;


			cout << "Statistics for level " << j << ":\n";
			Result.Print();
	
	
			for (i = 0; i < n; i++)
			{
				delete i_queues[i];
				i_queues[i] = NULL;
			}
			delete [] i_queues;
	


			delete imj_join_level[j];
			imj_join_level[j] = NULL;
	
			delete join_level[j];
			join_level[j] = NULL;


			if (Result.numRes <= 0)
			{
				printf("FULL INMEMORYJOIN WITH 0 TUPLES !!! EXIT !!!\n");
				exit(0);
			}
		}


		else
		{

			/* Now we load all of the non-constant guys into a priority Q for records: */
			//this data structure is used to sort tuples on TS so they can be fed into the IMJ in order
			//initialize the level input queues
			RecPriorityQ *outQ = new (std::nothrow) RecPriorityQ(n, isConstant);
			if (outQ == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
		
			for (i = 0; i < n; i++)
			{
				// only do this for the non-constant relations!!
				if (isConstant[i] == 0)
				{
					// just add all of them into the priority Q
					Record *rec = new (std::nothrow) Record();
					if (rec == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
	
					int numTuples = i_queues[i]->GetNumTuples();
//					numTuples = (numTuples > TUPLES_TO_IMJ) ? TUPLES_TO_IMJ : numTuples;
	
					for (k = 0; k < numTuples; k++) {
						i_queues[i]->GetNextTuple(rec, timeStamp);

//						cout << "Insert into outQ: " << timeStamp << "\n";
						outQ->Insert(rec, i, timeStamp);
					}

					samplingRatio[i] = (double)timeStamp / (double)HASH_DOMAIN;

					tuples_no[j][i] += numTuples;
	
					printf("Level %3d, Join %3d : INSERT INITIAL %20d IN PRIORITY QUEUE !!!\n", j, i, numTuples);
	
					delete rec;
				}
		
			}
	
			delete [] isConstant;
	
	
			/* Now we can load all of those initial records into the IMJ */
			// outQ will keep returning 1 as long as there is no input pipe that has been exhausted
			int whichRel;
	
			Record *rec = new (std::nothrow) Record();
			if (rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
	
//			cout << "About removing in outQ.\n";
			while (outQ->Remove(rec, timeStamp, whichRel)) {
				imj_join_level[j]->AddStartTuples(*rec, timeStamp, whichRel);
			}
			imj_join_level[j]->AddStartTuples(*rec, timeStamp, whichRel);
	
	
//			printf("The permutation order is to be determined.\n");
	
			// at this point, the IMJ is ready to go off and choose a permutation order
			SearchResults *myResult = imj_join_level[j]->ChoosePermOrder(samplingRatio);
//			SearchResults *myResult = imj_join_level[j]->ChoosePermOrder();

			printf("The permutation order was determined. Starting the normal execution !!!\n");
	
			if (myResult != NULL)
			{
				threshold = (double)timeStamp / (double)HASH_DOMAIN;
				stat_module->ProcessTuples(myResult, threshold);
	
				Result.numRes += myResult->numRes;
				Result.numDel += myResult->numDel;

//				Result.Print();
	
				delete myResult;
			}
	

			delete samplingRatio;

	
			/* At this point, the startup phase is totally done with.  So here is the main loop: */
			while (1)
			{
				// at the top of the loop, it is always the case that one of the
				// relations has no representatives in the outQ. So we go to that
				// one and get more tuples
		
				// first, we check to see if that one is done
				if (i_queues[whichRel]->IsAllDone())
				{
					// just notify the outQ
					outQ->ThisGuyIsDone(whichRel);
					numDone += 1;
		
					printf("Level %3d, Join %3d : DONE !!!\n", j, whichRel);
	
					// if all of them are done, exit the loop
					if (numDone == n) {
						break;
					}
				}
				else
				{
					// otherwise, we take and add a bunch of tuples into the outQ
					int numTuples = i_queues[whichRel]->GetNumTuples ();
					if (numTuples == 0)
					{
						if (i_queues[whichRel]->GetNextTuple(rec, timeStamp))
						{
							outQ->Insert (rec, whichRel, timeStamp);
	
							tuples_no[j][whichRel] += 1;
						}
						else
						{
							outQ->ThisGuyIsDone (whichRel);
							numDone += 1;
		
							printf("Level %3d, Join %3d : DONE !!!\n", j, whichRel);
	
							if (numDone == n) {
								break;
							}
						}
					}
		
					else
					{
	//					printf("Level %3d, Join %3d : INSERT %20d IN PRIORITY QUEUE !!!\n", j, whichRel, numTuples);
						for (k = 0; k < numTuples; k++) {
							if (i_queues[whichRel]->GetNextTuple(rec, timeStamp))
							{
								outQ->Insert(rec, whichRel, timeStamp);
	
								tuples_no[j][whichRel] += 1;
							}
							else
							{
								outQ->ThisGuyIsDone(whichRel);
		
								printf("Level %3d, Join %3d : DONE !!!\n", j, whichRel);
		
								numDone += 1;
								if (numDone == n) {
									break;
								}
							}
						}
					}
				}
		
				// now, we remove stuff from the outQ and add to the IMJ; take the results
				// and shuttle them to the statistics module
				while (outQ->Remove(rec, timeStamp, whichRel))
				{

					myResult = imj_join_level[j]->ExternInsertRec(*rec, whichRel, timeStamp);
	
					if (myResult != NULL)
					{
						threshold = (double)timeStamp / (double)HASH_DOMAIN;
						stat_module->ProcessTuples(myResult, threshold);
	
	
//						printf("Level %3d, Join %3d : INSERT !!!\n", j, whichRel);
	
//						printf("Level %3d, Join %3d : INSERT RESULT !!!\n", j, whichRel);
//						myResult->Print();

	
						Result.numRes += myResult->numRes;
						Result.numDel += myResult->numDel;
	
	
						delete myResult;
					}

				}
	
				myResult = imj_join_level[j]->ExternInsertRec(*rec, whichRel, timeStamp);
				if (myResult != NULL)
				{
					threshold = (double)timeStamp / (double)HASH_DOMAIN;
					stat_module->ProcessTuples(myResult, threshold);
	
	/*
					printf("Level %3d, Join %3d : INSERT !!!\n", j, whichRel);
	
					printf("Level %3d, Join %3d : INSERT RESULT !!!\n", j, whichRel);
					myResult->Print();
	*/
	
					Result.numRes += myResult->numRes;
					Result.numDel += myResult->numDel;
	
	
					delete myResult;
				}


//				Result.Print();
	
		
				// check to see if it is time to give out another result
				gettimeofday(&stat_current_time, NULL);
				unsigned int t_time = stat_current_time.tv_sec - stat_last_time.tv_sec;
				if (t_time >= UPDATE_FREQ)
				{
					// write the result to the file!
					stat_last_time = stat_current_time;
	
					threshold = (double)timeStamp / (double)HASH_DOMAIN;

					int estimates_no;
					double *est, *var, *low, *high;
					Record *groups_id;
		
					stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level, threshold);
	
					t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

					if (groups_no == -1)
					{
						printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", j, t_time, *low, *est, *high, *var, threshold);
					}
					else
					{
						for (i = 0; i < estimates_no; i++)
						{
							char *group_info;
							groups_id[i].Print(group_schema, group_info);
							printf("%s\n", group_info);

							printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", j, t_time, low[i], est[i], high[i], var[i], threshold);

							delete [] group_info;
						}
					}
		
					delete [] est;
					delete [] var;
					delete [] low;
					delete [] high;
					delete [] groups_id;
				}
			}
	
	
			cout << "Statistics for level " << j << ":\n";
			Result.Print();
	
	
	
			delete rec;
	
			delete outQ;
	
			for (i = 0; i < n; i++)
			{
				delete i_queues[i];
				i_queues[i] = NULL;

				printf("Level %3d, Join %3d : %20.2f tuples\n", j, i, tuples_no[j][i]);
			}
			delete [] i_queues;
	

	
			delete imj_join_level[j];
			imj_join_level[j] = NULL;
	
			delete join_level[j];
			join_level[j] = NULL;

		}


		//get the last estimates for the current level
		int estimates_no;
		double *est, *var, *low, *high;
		Record *groups_id;

		stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level, 1.0);

		gettimeofday(&stat_current_time, NULL);
		unsigned int t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

		cout << "End Level: " << j << "\n";

		if (groups_no == -1)
		{
			printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", j, t_time, *low, *est, *high, *var, 1.0);
		}
		else
		{
			for (i = 0; i < estimates_no; i++)
			{
				char *group_info;
				groups_id[i].Print(group_schema, group_info);
				printf("%s\n", group_info);

				printf("STATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", j, t_time, low[i], est[i], high[i], var[i], 1.0);

				delete [] group_info;
			}
		}

		delete [] est;
		delete [] var;
		delete [] low;
		delete [] high;
		delete [] groups_id;
	}
}


void ControllerTurbo::Control_UpJoinLevel()
{
	int i, n;

	int output_size;

	n = join_level[join_levels_no - 1]->GetSize();

	double *p = new (std::nothrow) double[n];
	if (p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	tuples_no[join_levels_no - 1] = new (std::nothrow) double[n];
	if (tuples_no[join_levels_no - 1] == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *cur_tot = new (std::nothrow) double[n];
	if (cur_tot == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (i = 0; i < n; i++)
	{
		p[i] = -1.0;
		tuples_no[join_levels_no - 1][i] = 0;
	}



	//add this level to the statistics module
	stat_module->AddGeneralLevel(n, 1);

	double tot_tuples = 0.0;

	while (1)
	{
		int still_running = 0;


		for (i = 0; i < n; i++)
		{
			int buffer_size = join_level[join_levels_no - 1]->out_pipe_sizes[i];

			Record *match_list = new (std::nothrow) Record[buffer_size];
			if (match_list == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			double in_l = join_level[join_levels_no - 1]->HowMuchIsLeft(i);

			int ret = join_level[join_levels_no - 1]->GetNextBatch(i, p[i], match_list, buffer_size, output_size);
			tuples_no[join_levels_no - 1][i] += output_size;

			if (ret == 0)
			{
				delete [] match_list;
				continue;
			}
			else
			{
				still_running = 1;
			}


			//build the aggregate function values for the returned tuples
			double *aggregate = new (std::nothrow) double[output_size];
			if (aggregate == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			Function *func = aggregate_level->GetFunction(0);
			for (int j = 0; j < output_size; j++)
			{
				int intResult;
				double doubleResult;
		
				Type t = func->Apply(&match_list[j], intResult, doubleResult);
		
				if (t == Int)
					aggregate[j] = (double)intResult;
				else if (t == Double)
					aggregate[j] = doubleResult;
			}


//			cout << "Tuples returned: " << output_size << "\n";



			//collapsing goes in here
			Schema *agg_schema = aggregate_level->GetSchema(0);

			int *join_atts;
			int join_atts_no;
			aggregate_level->GetJoinAtts(0, join_atts, join_atts_no);

			OrderMaker *join_order = new (std::nothrow) OrderMaker(agg_schema, join_atts, join_atts_no);
			if (join_order == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			int tup_no;

			if (groups_no == -1)
			{
				tup_no = Collaps_Aggregate_Level(match_list, aggregate, output_size, join_order);
//				int tup_no = Collaps_Aggregate_Level(match_list, aggregate, output_size);
//				int tup_no = output_size;
				tot_tuples += tup_no;
	
				delete join_order;
			}
			else
			{
				//in this case we have grouping
				//the grouping attributes have to be added to the OrderMaker
				int *group_atts;
				int group_atts_no;
				aggregate_level->GetGroupAtts(0, group_atts, group_atts_no);
	
				OrderMaker *group_order = new (std::nothrow) OrderMaker(agg_schema, group_atts, group_atts_no);
				if (group_order == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				OrderMaker *mixed_order = new (std::nothrow) OrderMaker();
				if (mixed_order == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				mixed_order->ANDMerge(join_order, group_order);

				delete group_order;
				delete join_order;

				tup_no = Collaps_Group_Level(match_list, aggregate, output_size, mixed_order);
				tot_tuples += tup_no;

				for (int j = 0; j < tup_no; j++)
				{
					match_list[j].Project(group_atts, group_atts_no, agg_schema->GetNumAtts());
				}

				delete mixed_order;
			}


//			cout << "Tuples collapsed: " << tup_no << "\n";


			cur_tot[i] = 1.0 - join_level[join_levels_no - 1]->HowMuchIsLeft(i);
//			printf("UpLevel, Join %3d : left %20.18f !!!\n", i, 1.0 - cur_tot[i]);


//statistics stuff gets in here
			SimpleSamplingData *sampling_prop = new (std::nothrow) SimpleSamplingData(n, cur_tot);
			if (sampling_prop == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			if (groups_no == -1)
			{
				stat_module->ProcessTuples(aggregate, NULL, tup_no, sampling_prop);
			}
			else
			{
				stat_module->ProcessTuples(aggregate, match_list, tup_no, sampling_prop);
			}

			delete sampling_prop;
//statistics stuff end


/*
			//the new collapsing code goes in here
			//first build the collaps structure
			SearchResults *myResult = new (std::nothrow) SearchResults;
			if (myResult == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			myResult->numConst = 1;

			myResult->constRels = new int;
			*myResult->constRels = 0;

			myResult->ratios = new double;
			*myResult->ratios = 1.0;

			myResult->numDel = 0;
			myResult->funcDel = NULL;
			myResult->delTuples = NULL;

			myResult->numRes = output_size;
			myResult->funcRes = aggregate;

			myResult->resTuples = new (std::nothrow) int*[myResult->numRes];
			if (myResult->resTuples == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			for(int j = 0; j < myResult->numRes; j++)
			{
				myResult->resTuples[j] = new int;
				*myResult->resTuples[j] = match_list[j].hash;
			}


			cur_tot[i] = 1.0 - join_level[join_levels_no - 1]->HowMuchIsLeft(i);
//			printf("UpLevel, Join %3d : left %20.18f !!!\n", i, 1.0 - cur_tot[i]);

			stat_module->ProcessTuples(myResult, cur_tot[0]);

			delete myResult;
*/



			delete [] match_list;
			delete [] aggregate;
		}

		// check to see if it is time to give out another result
		gettimeofday(&stat_current_time, NULL);
		unsigned int t_time = stat_current_time.tv_sec - stat_last_time.tv_sec;
		if (t_time >= UPDATE_FREQ)
		{
			// write the result to the file!
			stat_last_time = stat_current_time;

			int estimates_no;
			double *est, *var, *low, *high;
			Record *groups_id;
	
			stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level);

			t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

			if (groups_no == -1)
			{
				printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, *low, *est, *high, *var, cur_tot[0]);
			}
			else
			{
				for (i = 0; i < estimates_no; i++)
				{
					char *group_info;
					groups_id[i].Print(group_schema, group_info);
					printf("%s\n", group_info);

					printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, low[i], est[i], high[i], var[i], cur_tot[0]);

					delete [] group_info;
				}
			}
	
			delete [] est;
			delete [] var;
			delete [] low;
			delete [] high;
			delete [] groups_id;
		}

		if (still_running == 0)
			break;
	}

	delete [] p;
	delete [] cur_tot;


//last call for estimates before exiting
	int estimates_no;
	double *est, *var, *low, *high;
	Record *groups_id;

	stat_module->GetEst(est, var, low, high, groups_id, estimates_no, conf_level);

	gettimeofday(&stat_current_time, NULL);
	unsigned int t_time = stat_current_time.tv_sec - stat_start_time.tv_sec;

	cout << "End Level: " << join_levels_no - 1 << "\n";

	if (groups_no == -1)
	{
		printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, *low, *est, *high, *var, 1.0);
	}
	else
	{
		for (i = 0; i < estimates_no; i++)
		{
			char *group_info;
			groups_id[i].Print(group_schema, group_info);
			printf("%s\n", group_info);

			printf("UPSTATS %3d %10d : %30.4f\t %30.4f\t %30.4f\t %40.10f %7.5f\n", join_levels_no - 1, t_time, low[i], est[i], high[i], var[i], 1.0);

			delete [] group_info;
		}
	}

	delete [] est;
	delete [] var;
	delete [] low;
	delete [] high;
	delete [] groups_id;


	for (int jj = 0; jj < join_level[join_levels_no - 1]->GetSize(); jj++)
	{
		printf("Level %3d, Join %3d : %20.2f tuples\n", join_levels_no - 1, jj, tuples_no[join_levels_no - 1][jj]);
	}


//	cout << "Total Tuples : " << tot_tuples << "\n";


	delete join_level[join_levels_no - 1];
	join_level[join_levels_no - 1] = NULL;

}


void ControllerTurbo::Control_AggregateLevel()
{
	int i, n;

	n = aggregate_level->GetSize();
	for (i = 0; i < n; i++)
	{
		//fake call to GetNextBatch for synchronization purposes
		int output_size;
		int ret = aggregate_level->GetNextBatch(i, 0.0, NULL, -1, output_size);

		double aggregate = aggregate_level->GetAggregate(i);
		double tup_no = aggregate_level->GetTuplesNo(i);

		printf("Result: %30.4f\n", aggregate);
		printf("Tuples: %30.1f\n\n", tup_no);
	}
}


void ControllerTurbo::Control_GroupAggregateLevel()
{
	int i, n;

	n = aggregate_level->GetSize();
	for (i = 0; i < n; i++)
	{
		//fake call to GetNextBatch for synchronization purposes
		int output_size;
		int ret = aggregate_level->GetNextBatch(i, 0.0, NULL, -1, output_size);

		Schema *schema = aggregate_level->GetSchema(i);

		int *group_atts;
		int group_atts_no;
		aggregate_level->GetGroupAtts(i, group_atts, group_atts_no);

		schema->Project(group_atts, group_atts_no);

		vector<Record*> desc;
		vector<double> aggregate;
		vector<double> tuples;

		int extract_groups_no;

		aggregate_level->GetGroups(i, desc, aggregate, tuples, extract_groups_no);

		for (int j = 0; j < extract_groups_no; j++)
		{
			desc[j]->Print(schema);
			printf("Result: %30.4f\n", aggregate[j]);
			printf("Tuples: %30.1f\n", tuples[j]);
		}

		printf("Groups No: %30.1f\n\n", (double)extract_groups_no);
	}
}






//collapsing method for the aggregate level; it aggregates the function values for each of the distinct groups
int Collaps_Aggregate_Level(Record *join_res, double *&output, int v_size, OrderMaker *join_order)
//int Collaps_Aggregate_Level(Record *join_res, double *&output, int v_size)
{
	Hash *eq_func = new (std::nothrow) Hash();
	if (eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *new_output = new (std::nothrow) double[v_size];
	if (new_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	vector<int> index;
	vector<double> sum_func;

	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		index.push_back(c_i);
		sum_func.push_back(output[c_i]);

//		printf("HASH : %30d\n", join_res[c_i].hash);

		c_j = c_i + 1;
		while ((c_j < v_size) && (join_res[c_i].hash == join_res[c_j].hash))
		{
			int is_in = 0;
			int nn = index.size();

			for (int i = 0; i < nn; i++)
			{
				if (eq_func->Compare(&join_res[index[i]], &join_res[c_j], join_order) == 0)
				{
					//check to see if this record already exists
					sum_func[i] = sum_func[i] + output[c_j];

					is_in = 1;
					break;
				}
			}

			if (is_in == 0)
			{
				//in this case there is a conflict in the hash
				index.push_back(c_j);
				sum_func.push_back(output[c_j]);

//				cout << "Conflict!!!\n";
			}

			c_j += 1;
		}

		int nn = index.size();
		for (int i = 0; i < nn; i++)
		{
			new_output[colaps_count + i] = sum_func[i];
		}

		colaps_count += nn;

		index.clear();
		sum_func.clear();

		c_i = c_j;
	}


	delete [] output;
	output = new_output;

	delete eq_func;

	return colaps_count;



/*
	Hash *eq_func = new (std::nothrow) Hash();
	if (eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *new_output = new (std::nothrow) double[v_size];
	if (new_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		new_output[colaps_count] = output[c_i];

		c_j = c_i + 1;


		while ((c_j < v_size) && (eq_func->Compare(&join_res[c_i], &join_res[c_j], join_order) == 0))
		{
			new_output[colaps_count] += output[c_j];

			c_j += 1;
		}

		c_i = c_j;
		colaps_count += 1;
	}


	delete [] output;
	output = new_output;

	delete eq_func;

	return colaps_count;
*/

}




//collapsing method for the aggregate level; it aggregates the function values for each of the distinct groups
int Collaps_Group_Level(Record *&join_res, double *&output, int v_size, OrderMaker *join_order)
{
	Hash *eq_func = new (std::nothrow) Hash();
	if (eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *new_output = new (std::nothrow) double[v_size];
	if (new_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Record *new_join_res = new (std::nothrow) Record[v_size];
	if (new_join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	vector<int> index;
	vector<double> sum_func;

	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		index.push_back(c_i);
		sum_func.push_back(output[c_i]);

//		printf("HASH : %30d\n", join_res[c_i].hash);

		c_j = c_i + 1;
		while ((c_j < v_size) && (join_res[c_i].hash == join_res[c_j].hash))
		{
			int is_in = 0;
			int nn = index.size();

			for (int i = 0; i < nn; i++)
			{
				if (eq_func->Compare(&join_res[index[i]], &join_res[c_j], join_order) == 0)
				{
					//check to see if this record already exists
					sum_func[i] = sum_func[i] + output[c_j];

					is_in = 1;
					break;
				}
			}

			if (is_in == 0)
			{
				//in this case there is a conflict in the hash
				index.push_back(c_j);
				sum_func.push_back(output[c_j]);

//				cout << "Conflict!!!\n";
			}

			c_j += 1;
		}

		int nn = index.size();
		for (int i = 0; i < nn; i++)
		{
			new_output[colaps_count + i] = sum_func[i];
			new_join_res[colaps_count + i].Consume(&join_res[index[i]]);
		}

		colaps_count += nn;

		index.clear();
		sum_func.clear();

		c_i = c_j;
	}


	delete [] output;
	output = new_output;

	delete [] join_res;
	join_res = new_join_res;

	delete eq_func;

	return colaps_count;
}



//collapsing method based on linear hashing; it aggregates the function values for each of the distinct groups
int Collaps_Hash(int **&join_res, double *&output, int v_size, int relations_no)
{

	//THIS IMPLEMENTATION IS BASED ON HASHING//

	int *colaps_array = new (std::nothrow) int[v_size];
	if (colaps_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < v_size; i++)
	{
		colaps_array[i] = -1;
	}

	int colaps_count = 0;
	for (int i = 0; i < v_size; i++)
	{
		//compute the index in the hash array of each record in the output

		unsigned int index = 0;
		for (int j = 0; j < relations_no; j++)
		{
			index += join_res[i][j];

		}
		index = index % v_size;

		if (colaps_array[index] == -1)
		{
			//in this case the position is free, so just add the index to the array
			colaps_array[index] = i;
			colaps_count += 1;

		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = memcmp(join_res[i], join_res[colaps_array[index]], relations_no * sizeof(int));
			if (are_identical == 0)
			{
				//the records are identical, so collapse the aggregate values
				output[colaps_array[index]] += output[i];

			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % v_size; (colaps_array[j] != -1) && (j != index); j = (j + 1) % v_size)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = memcmp(join_res[i], join_res[colaps_array[j]], relations_no * sizeof(int));
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					output[colaps_array[j]] += output[i];
				}
				else
				{
					if (colaps_array[j] == -1)
					{
						colaps_array[j] = i;
						colaps_count += 1;
					}
					else
					{
						cout << "This should be an error. No place was found in the hash array for record with index " << i << " in the result.";
						exit(1);
					}
				}
			}
		}
	}

	int **p_join_res = new (std::nothrow) int*[colaps_count];
	if (p_join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *p_output = new (std::nothrow) double[colaps_count];
	if (p_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

//	double tot_res = 0.0;

	int c_i = 0;
	for (int i = 0; i < v_size; i++)
	{
		if (colaps_array[i] != -1)
		{
			p_join_res[c_i] = new (std::nothrow) int[relations_no];
			if (p_join_res[c_i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			memcpy(p_join_res[c_i], join_res[colaps_array[i]], relations_no * sizeof(int));
	
			p_output[c_i] = output[colaps_array[i]];

//			tot_res += p_output[c_i];

			c_i += 1;
			if (c_i > colaps_count)
			{
				cout << "Error with putting the collapsed records together !";
				exit(1);
			}
		}

		delete [] join_res[i];
	}

	delete [] join_res;
	join_res = p_join_res;

	delete [] output;
	output = p_output;


	delete [] colaps_array;

	//end of collapsing


// 	printf("InMemoryJoinResult: %30.5f\n\n", tot_res);


	return colaps_count;
}



//collapsing method based on sorting; it aggregates the function values for each of the distinct groups
int Collaps_Sort(int **&join_res, double *&output, int v_size, int relations_no)
{
	//THIS IMPLEMENTATION IS BASED ON SORTING//

	int *colaps_array = new (std::nothrow) int[v_size * (relations_no + 2)];
	if (colaps_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < v_size; i++)
	{
		colaps_array[i * (relations_no + 2)] = relations_no;

		memcpy(colaps_array + i * (relations_no + 2) + 1, join_res[i], relations_no * sizeof(int));

		colaps_array[i * (relations_no + 2) + relations_no + 1] = i;

		delete [] join_res[i];
	}

	delete [] join_res;

	qsort(colaps_array, v_size, sizeof(int) * (relations_no + 2), Compare_Direct_IMJ);

	int *colaps_index = new (std::nothrow) int[v_size];
	if (colaps_index == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		colaps_index[colaps_count] = c_i;
		colaps_count += 1;

		c_j = c_i + 1;


		while ((c_j < v_size) && (memcmp(colaps_array + c_i * (relations_no + 2), colaps_array + c_j * (relations_no + 2), (relations_no + 1) * sizeof(int)) == 0))
		{
			output[colaps_array[c_i * (relations_no + 2) + relations_no + 1]] += output[colaps_array[c_j * (relations_no + 2) + relations_no + 1]];

			c_j += 1;
		}

		c_i = c_j;
	}


	//here we have in colaps_index colaps_count records that have survived

	join_res = new (std::nothrow) int*[colaps_count];
	if (join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *p_output = new (std::nothrow) double[colaps_count];
	if (p_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


//	double tot_res = 0.0;


	for (int i = 0; i < colaps_count; i++)
	{
		join_res[i] = new (std::nothrow) int[relations_no];
		if (join_res[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		memcpy(join_res[i], colaps_array + colaps_index[i] * (relations_no + 2) + 1, relations_no * sizeof(int));

		p_output[i] = output[colaps_array[colaps_index[i] * (relations_no + 2) + relations_no + 1]];

//		tot_res += p_output[i];
	}
	//done with putting the results in the output


	delete [] colaps_index;

	delete [] colaps_array;

	delete [] output;
	output = p_output;

	//end of collapsing

// 	printf("InMemoryJoinResult: %30.5f\n\n", tot_res);


	return colaps_count;
}



int Compare_IMJ(const void *r1, const void *r2)
{
	int *a, *b;

	a = (int*)*(int*)r1;
	b = (int*)*(int*)r2;

	for (int i = 1; i <= *a; i++)
	{
		if (*(a + i) > *(b + i))
			return 1;
		else if (*(a + i) < *(b + i))
			return -1;
	}

	return 0;
}


int Compare_Direct_IMJ(const void *r1, const void *r2)
{
	int *a, *b;

	a = (int*)r1;
	b = (int*)r2;

	int res = memcmp(a + 1, b + 1, *a * sizeof(int));
	return res;

/*
	for (int i = 1; i <= *a; i++)
	{
		if (*(a + i) > *(b + i))
			return 1;
		else if (*(a + i) < *(b + i))
			return -1;
	}

	return 0;
*/

}

