#ifndef FILE_H
#define FILE_H

#include "Hashes.h"
#include "TwoWayList.h"
#include "Record.h"

#include <vector>
#include <deque>
#include <pthread.h>

//this is used for striping
#define NUM_DISKS 4

#define DONE 0
#define READ 1
#define WRITE 2


//this is used for C-STORE
#define FT_NONE 0
#define FT_INT 1
#define FT_REAL 2
#define FT_STRING 3
#define FT_RECORD 4



using namespace std;



struct SortRecord {
	Record *record;
	OrderMaker *order;
	Hash *hash;
};


struct ThreadParam {
  char **bits;

  int *status;
  int *execute;
  int *counter;

  int *myFD;

  pthread_mutex_t *mainMutex;
  pthread_mutex_t *myMutex;

  pthread_cond_t *mainSignal;
  pthread_cond_t *myCond;
};




class Page {
private:
	TwoWayList <Record> *myRecs;
	
	int numRecs;
	int curSizeInBytes;

public:
	// constructor
	Page ();
	virtual ~Page ();

	// this takes a page and writes its binary representation to bits
	void ToBinary (char *bits);

	// this takes a binary representation of a page and gets the
	// records from it
	void FromBinary (char *bits);

	// the deletes the first record from a page and returns it; returns
	// a zero if there were no records on the page
	int GetFirst (Record *firstOne);

	// this appends the record to the end of a page.  The return value
	// is a one on success and a aero if there is no more space
	// note that the record is consumed so it will have no value after
	int Append (Record *addMe);

	// empty it out
	void EmptyItOut ();

};



class C_Page {
protected:
	int numItems;
	int curSizeInBytes;

	deque<char*> *elem;

public:
	// constructor
	C_Page ();
	virtual ~C_Page ();

	// this takes a page and writes its binary representation to bits
	virtual void ToBinary (char *bits) = 0;

	// this takes a binary representation of a page and gets the
	// records from it
	virtual void FromBinary (char *bits) = 0;

	// the deletes the first record from a page and returns it; returns
	// a zero if there were no records on the page
	virtual int GetFirst (char *&first) = 0;

	// this appends the record to the end of a page.  The return value
	// is a one on success and a aero if there is no more space
	// note that the record is consumed so it will have no value after
	virtual int Append (char *&addMe) = 0;

	// empty it out
	virtual void EmptyItOut () = 0;
};


class C_Page_INT : public C_Page {
public:
	// constructor
	C_Page_INT ();
	virtual ~C_Page_INT ();

	// this takes a page and writes its binary representation to bits
	virtual void ToBinary (char *bits);

	// this takes a binary representation of a page and gets the
	// records from it
	virtual void FromBinary (char *bits);

	// the deletes the first record from a page and returns it; returns
	// a zero if there were no records on the page
	virtual int GetFirst (char *&first);

	// this appends the record to the end of a page.  The return value
	// is a one on success and a aero if there is no more space
	// note that the record is consumed so it will have no value after
	virtual int Append (char *&addMe);

	// empty it out
	virtual void EmptyItOut ();
};


class C_Page_DOUBLE : public C_Page {
public:
	// constructor
	C_Page_DOUBLE ();
	virtual ~C_Page_DOUBLE ();

	// this takes a page and writes its binary representation to bits
	virtual void ToBinary (char *bits);

	// this takes a binary representation of a page and gets the
	// records from it
	virtual void FromBinary (char *bits);

	// the deletes the first record from a page and returns it; returns
	// a zero if there were no records on the page
	virtual int GetFirst (char *&first);

	// this appends the record to the end of a page.  The return value
	// is a one on success and a aero if there is no more space
	// note that the record is consumed so it will have no value after
	virtual int Append (char *&addMe);

	// empty it out
	virtual void EmptyItOut ();
};


class C_Page_STRING : public C_Page {
protected:
	deque<int> *e_size;

public:
	// constructor
	C_Page_STRING ();
	virtual ~C_Page_STRING ();

	// this takes a page and writes its binary representation to bits
	virtual void ToBinary (char *bits);

	// this takes a binary representation of a page and gets the
	// records from it
	virtual void FromBinary (char *bits);

	// the deletes the first record from a page and returns it; returns
	// a zero if there were no records on the page
	virtual int GetFirst (char *&first);

	// this appends the record to the end of a page.  The return value
	// is a one on success and a aero if there is no more space
	// note that the record is consumed so it will have no value after
	virtual int Append (char *&addMe);

	// empty it out
	virtual void EmptyItOut ();
};


class C_Page_RECORD : public C_Page {
protected:
	deque<int> *e_size;

public:
	// constructor
	C_Page_RECORD ();
	virtual ~C_Page_RECORD ();

	// this takes a page and writes its binary representation to bits
	virtual void ToBinary (char *bits);

	// this takes a binary representation of a page and gets the
	// records from it
	virtual void FromBinary (char *bits);

	// the deletes the first record from a page and returns it; returns
	// a zero if there were no records on the page
	virtual int GetFirst (char *&first);

	// this appends the record to the end of a page.  The return value
	// is a one on success and a aero if there is no more space
	// note that the record is consumed so it will have no value after
	virtual int Append (char *&addMe);

	// empty it out
	virtual void EmptyItOut ();
};






class File_Interface {
protected:
	C_Page *myBuff;

	off_t curLength;
	off_t curPos;
	int changed;

	int *myFilDes;

	char *path;

	int file_type;

protected:
	/** Changes the file position. */
	virtual void Lseek (off_t howFar) = 0;

	/** Writes an array of characters to disk. */
	virtual void Write (char *writeMe) = 0;

	/** Reads from an array of characters. */
	virtual void Read (char *readMe) = 0;

public:
	File_Interface();
	virtual ~File_Interface();


	// returns the current length of the file, in pages
	virtual off_t GetLength ();


	// moves the record "pointer" in the file to the first record on 
	// the specified page in the file
	virtual void MovePointer (off_t whichPage);


	// gets the next record in the file; returns a 0 if there are no more
	// records and a 1 otherwise
	virtual int NextRec (Record *getMe);

	// adds a new record to the very end of the file
	virtual void Append (Record *addMe);


	// allows someone to explicitly get a specified page from the file
	virtual void GetPage (Page *putItHere, off_t whichPage);

	// allows someone to explicitly write a specified page to the file
	// if the write is past the end of the file, all of the new pages that
	// are before the page to be written are zeroed out
	virtual void AddPage (Page *addMe, off_t whichPage);


	// opens the given file; the first parameter tells whether or not to
	// create the file.  If the parameter is zero, a new file is created
	// the file; if notNew is zero, then the file is created and any other
	// file located at that location is erased.  Otherwise, the file is
	// simply opened
	virtual void Open (int notNew, char *fName, int file_type) = 0;

	// closes the file and returns the file length (in number of pages)
	virtual int Close () = 0;

	// closes the file and removes it from disk
	virtual void Kill () = 0;

};




class File : public File_Interface {
protected:
	/** Changes the file position. */
	virtual void Lseek (off_t howFar);

	/** Writes an array of characters to disk. */
	virtual void Write (char *writeMe);

	/** Reads from an array of characters. */
	virtual void Read (char *readMe);

public:
	File ();
	virtual ~File ();


	// opens the given file; the first parameter tells whether or not to
	// create the file.  If the parameter is zero, a new file is created
	// the file; if notNew is zero, then the file is created and any other
	// file located at that location is erased.  Otherwise, the file is
	// simply opened
	virtual void Open (int notNew, char *fName, int file_type);

	// closes the file and returns the file length (in number of pages)
	virtual int Close ();

	// closes the file and removes it from disk
	virtual void Kill ();

};



/*
class that stores a file divided over multiple disks
*/
class Stripe_File : public File_Interface {
protected:

	/** Set of mutexes. */
	pthread_mutex_t mutexes[NUM_DISKS];
	/** Set of condition signals. */
	pthread_cond_t mySignals[NUM_DISKS];

	/** Main mutexes. */
	pthread_mutex_t mainMutex;
	/** Main condition signal. */
	pthread_cond_t mainSignal;
	

	/** Set of threads. */
	pthread_t myThreads[NUM_DISKS];

	/** Thread parameters. */
	ThreadParam *myThreadParam;


	/** Status array for every disk. */
	int *status;

	/** Execute array for every disk. */
	int *execute;

	/** Counter. */
	int *counter;

	/** Bit pointers for every disk. */
	char **bitPointers;


protected:
	/** Changes the file position. */
	virtual void Lseek (off_t howFar);

	/** Writes an array of characters to disk. */
	virtual void Write (char *writeMe);

	/** Reads from an array of characters. */
	virtual void Read (char *readMe);

public:
	/** Simple constructor. */
	Stripe_File ();
	
	/** Simple deconstructor. */
	virtual ~Stripe_File ();


	/** Opens the given file.
	*
	*  @param notNew Tells whether or not to
	*         create the file.  If it is zero, a new file is created.
	*         Otherwise, the file is simply opened.
	*/
	virtual void Open (int notNew, char *fName, int file_type);

	/** Closes the file and returns the file length (in number of pages). */
	virtual int Close ();
	
	/** Closes the file and removes it from disk. */
	virtual void Kill ();
};





/*
class used for storing all the runs of a sorted relation in a single file
if multiple files are used (one for each run), it could be the case that the number of file descriptors available in the OS is insufficent
*/
class Run_File_Interface {
protected:
	//one page for each run
	Page **myBuff;
	off_t runs_no;

	//the total length of the file in pages
	off_t current_length;

	//run length in records
	vector<off_t> run_length;
	//run start page
	vector<off_t> run_start_page;
	//current record number for each run
	vector<off_t> cur_pos;
	//current page position for each run
	vector<off_t> cur_page_pos;

	int *myFilDes;

	char *path;

protected:
	/** Changes the file position. */
	virtual void Lseek (off_t howFar) = 0;

	/** Writes an array of characters to disk. */
	virtual void Write (char *writeMe) = 0;

	/** Reads from an array of characters. */
	virtual void Read (char *readMe) = 0;


public:
	Run_File_Interface ();
	virtual ~Run_File_Interface ();


	// opens the given file; the first parameter tells whether or not to create the file.  If the parameter is zero, a new file is created the file; if notNew is zero, then the file is created and any other file located at that location is erased.  Otherwise, the file is simply opened
	virtual void Open (int notNew, char *fName) = 0;

	// closes the file and returns the file length (in number of pages); if write_needed is not 0, the current pages in the runs need to be written; if write_needed is 0, the pages are not written back in the file
	virtual int Close (int write_needed) = 0;

	// closes the file and removes it from disk
	virtual void Kill () = 0;


	// moves the record "pointer" corresponding to run_no to the first record on the next page of this run
	virtual void MovePointer (int run_no);

	// gets the next record in the file corresponding to run number run_no; returns a 0 if there are no more records and a 1 otherwise
	virtual int NextRec (Record *getMe, int run_no);

	// append a new run to the file; the records of the run are stored sorted in SortRecord which has the size buffer_size; return the new size of the file on success, or -1 on failure
	virtual off_t Append_Run(SortRecord *buffer, int buffer_size);
};



/*
class used for storing all the runs of a sorted relation in a single file
if multiple files are used (one for each run), it could be the case that the number of file descriptors available in the OS is insufficent
*/
class Run_File : public Run_File_Interface {
protected:
	/** Changes the file position. */
	virtual void Lseek (off_t howFar);

	/** Writes an array of characters to disk. */
	virtual void Write (char *writeMe);

	/** Reads from an array of characters. */
	virtual void Read (char *readMe);

public:
	Run_File ();
	virtual ~Run_File ();

	// opens the given file; the first parameter tells whether or not to create the file.  If the parameter is zero, a new file is created the file; if notNew is zero, then the file is created and any other file located at that location is erased.  Otherwise, the file is simply opened
	virtual void Open (int notNew, char *fName);

	// closes the file and returns the file length (in number of pages); if write_needed is not 0, the current pages in the runs need to be written; if write_needed is 0, the pages are not written back in the file
	virtual int Close (int write_needed);

	// closes the file and removes it from disk
	virtual void Kill ();
};




class Run_Stripe_File : public Run_File_Interface {
protected:
	/** Set of mutexes. */
	pthread_mutex_t mutexes[NUM_DISKS];
	/** Set of condition signals. */
	pthread_cond_t mySignals[NUM_DISKS];

	/** Main mutexes. */
	pthread_mutex_t mainMutex;
	/** Main condition signal. */
	pthread_cond_t mainSignal;
	

	/** Set of threads. */
	pthread_t myThreads[NUM_DISKS];

	/** Thread parameters. */
	ThreadParam myThreadParam[NUM_DISKS];

	/** Status array for every disk. */
	int status[NUM_DISKS];

	/** Execute array for every disk. */
	int execute[NUM_DISKS];

	/** Counter. */
	int *counter;

	/** Bit pointers for every disk. */
	char *bitPointers[NUM_DISKS];

protected:
	/** Changes the file position. */
	virtual void Lseek (off_t howFar);

	/** Writes an array of characters to disk. */
	virtual void Write (char *writeMe);

	/** Reads from an array of characters. */
	virtual void Read (char *readMe);


public:
	Run_Stripe_File ();
	virtual ~Run_Stripe_File ();


	// opens the given file; the first parameter tells whether or not to create the file.  If the parameter is zero, a new file is created the file; if notNew is zero, then the file is created and any other file located at that location is erased.  Otherwise, the file is simply opened
	virtual void Open (int notNew, char *fName);

	// closes the file and returns the file length (in number of pages); if write_needed is not 0, the current pages in the runs need to be written; if write_needed is 0, the pages are not written back in the file
	virtual int Close (int write_needed);

	// closes the file and removes it from disk
	virtual void Kill ();
};



//global thread function for stripping
void *Stripe_Thread (void *param);


#endif
