// The purpose of this class is to efficiently join a set of tuples that
// are all in memory.  The way that this will be sued in practice is that
// for a set of relations, a bunch of records from each relation will be
// loaded into the structure, and then the structure will be asked to compute
// a join over them.  Subsequently, the set of records from one of the relations
// will be removed (via a call to Swap) and replaced with a new set of records
// from the same relation.  Then, the strucutre will be asked to recompute the
// answer to the join.  Because of this, the structure should be clever and
// somehow save some of its intermediate computations.  For example, imagine
// that we have a join of R, S, T and U.  R and S are linked, S and T are
// linked, and T and U are linked.  If you replace the records that came from
// R, then this should not have any effect on the links from S to T and from
// T to U, so all of that information should not need to be recomputed!

#ifndef IN_MEMORY_JOIN_H
#define IN_MEMORY_JOIN_H


#include "Comparison.h"
#include "Function.h"
#include "Record.h"
#include "Hashes.h"




class InMemoryJoin {
	public:
		// the first param tells how many relations are taking place in the 
		// join.  The second parameter is a 2-D array of CNF expressions that
		// link up the input relations.  For example, myCNF[2][3] is the CNF
		// expression that is used to join the third and fourth input relations,
		// respecitvely.  mask tells which of the CNF expressions are actually
		// useful (since some of them might not apply to the query in question) 
		// if mask[i][j] is 1, then the corresponding CNF entry is used; if it
		// is zero, then the corresponding CNF entry is never used.  By convention,
		// myCNF[i][j] for i >= j is never used.  The myCNF array is consumed
		// by the constructor, so that after the call it should be 0.
		InMemoryJoin(int relations_no, CNF **cnf, CNF *extra_cond, Record *extra_lit, int **mask, Schema **relSchemas, Function *toApply);
		~InMemoryJoin();

		// this inserts the records from one of the input relations into the
		// structure.  The records are consumed, so that after the call, input
		// is 0.  If there are already in records inside of the structure for
		// that particular relation, then the ones that are already there are
		// returned to the caller in the output parameter (this parameter returns
		// a 0 if there were no records there).
		//
		// ** LL(1): I added the numRecords parameter. Represents the size of the input array and returns the size of the output array after the call is finished.
		void Swap(Record *&input, Record *&output, int which_relation, int &output_no);

		// this computes a join over all of the records that are currently 
		// in the structure, using the join predicates in myCNF.  The results
		// are loaded in to the array whichRecs.  If whichRecs[i][j] = k, then
		// it means that the ith record output from the join includes record k
		// from input relation j.  The return value from the function is the 
		// number of records included in the join.  Thus, j goes from 0 to 
		// numRelations, and k goes from 0 to the number of records from relation
		// j that are inside of the structure. If with_collaps = 1, the collapsing is applied since a variance computation needs to be done; if with_collaps = 0, no collapsing is applied.
		int Join(int **&join_res, double *&output);


	private:
		struct Pair
		{
			int c1;
			int c2;
		};

		struct HashChain
		{
			int rec_index;
			HashChain *next;
		};


		// destroys a given hash table
		void DestroyHashTable(int i, int j);

		// DFS path
		int GetDFSOrder(int k, Pair *&output);



	public:
		int relations_no;
		int *schemaSizes;

		Schema **schemas;

		OrderMaker **cnf;
		OrderMaker *colaps_om;

		CNF *extra_cond;
		Record *extra_lit;

		int **mask;

		Function *func;

		Record **records;		// every record so far
		int *recs_no;			// number of records in each entry of records[]

		HashChain ****hash_table; 	// the hash thing
		Hash ***hash_func;
		Hash *exact_hash;

		Hash **colaps_hash;


		Pair *dfs_order;
		int pairs_no;

};

#endif
