#ifndef GROUP_COMPILER_H
#define GROUP_COMPILER_H

#include "Pipe.h"
#include "ParseTree.h"
#include "Function.h"
#include "Levelwise.h"
#include "InMemoryJoin.h"
#include "NewInMemoryJoin.h"

#include <string>
#include <set>
#include <vector>
#include <iostream>
#include <map>

#define MAXLEVEL 10
#define PIPE_BUFFERSIZE 10000



using namespace std;


class OperationItem
{
public:
	int flag;
	
	int mylevel;		//this is used to specify the level of a node in a tree
				//it will be used to check whether those nodes claimed in the same levelwise step
				//are actually in the same leve in the query tree.
	int type;
	
	int InputPipe1;
	int InputPipe2;
	int OutputPipe;

	char *schema_file;
	Schema *InputSchema1, *InputSchema2, *OutputSchema;
	
	
	int *AttstoProj;
	int NumAttstoProj;
	
	int InMemoryJoinBufferSize;

	set<string> CurrentLevelAtts;
	
	OperationItem *parent, *child1, *child2; 


	OperationItem(char *schema_file);
	virtual ~OperationItem();

	void CheckCircle();

	int Count();

	void SetLevel(int i);



	virtual bool IsUnary()=0;

	virtual void BuildSchema(set<string> AttsFromAbove)=0;

	virtual void PrintParameters(int recursive)=0;
};




class TableScanOperation : public OperationItem
{
public:
	AndList *CurrentAndList;

	CNF *myCNF;

	Record *RecLit;
	Schema *RecLit_Schema;


	TableScanOperation(char *schema_file);
	virtual ~TableScanOperation();


	void Initialization(TableScanYac *CurrentOperation);
	

	virtual bool IsUnary();
	
	virtual void BuildSchema(set<string> AttsFromAbove);
	
	virtual void PrintParameters(int recursive);
};

class MultiScanOperation : public OperationItem
{
public:
	char **file_location;
	Schema **schema;

	int **proj_atts;
	int *proj_atts_no;

	int *proj_atts_out;
	int proj_atts_out_no;

	int *tuples_no;
	int relations_no;
	
	struct AttrGroupbyList *Relations;
	AndList *CurrentAndList;


	MultiScanOperation(char *schema_file);
	virtual ~MultiScanOperation();

	void Initialization(MultiScanYac *CurrentOperation);

	virtual bool IsUnary();
	virtual void BuildSchema(set<string> AttsFromAbove);
	virtual void PrintParameters(int recursive);
};


class SortJoinOperation : public OperationItem
{
public:
	int MemorySize;

	AndList *CurrentAndList;

	CNF *myCNF;

	Record *RecLit;
	Schema *RecLit_Schema;
	

	SortJoinOperation(char *schema_file);
	virtual ~SortJoinOperation();


	void Initialization(SortJoinYac *CurrentOperation);

	virtual bool IsUnary();

	virtual void BuildSchema(set<string> AttsFromAbove);

	virtual void PrintParameters(int recursive);
};


class RehashingOperation : public OperationItem
{
public:
	int MemorySize;
	RehashingOperation(char *schema_file);
	
	virtual ~RehashingOperation();
	
	void Initialization(RehashingYac *CurrentOperation);
	
	virtual bool IsUnary();

	virtual void BuildSchema(set<string> AttsFromAbove);

	virtual void PrintParameters(int recursive);
};



class AggregateOperation : public OperationItem
{
public:
	FuncOperator *CurrentOperator;
	Function *myFunc;
	set<string> JoinAtts;	
	int *JoinAttIndex;
	int numJoinAtts;	
	struct AttrGroupbyList *myGroupBy;
  	int numGroupstoDisplay;
	int queueSize;
  	int MemorySize;
  	int numTuplesforStats;
	int *GroupingIndex;
	int numGroupingIndex;

	AggregateOperation(char *schema_file);
	virtual ~AggregateOperation();


	void Initialization(AggregateStep *CurrentOperation);


	virtual bool IsUnary();

	virtual void BuildSchema(set<string> AttsFromAbove);

	virtual void PrintParameters(int recursive);


private:
	void PrepareAtts(FuncOperator *Operator);
};





class LevelWiseStepCompiler
{
public:
	int LevelWiseNumber;
	int softBound, hardBound;

        vector<OperationItem*> Operations;
	vector<int> scanType;

	bool LevelChecked;


	//the following are used for preparing in-memory-join parameters
	//remember we follow the rule that in-memory-join always takes sample from output pipe
	int ** InMemoryJoinMask;

	CNF ** InMemoryCNFs;
	CNF * OtherConditions;

	Function *InMemoryFunc;
	Schema *func_schema;

	Record *RecLit;
	Schema *RecLit_Schema;

	AndList* myList;
	
	LevelWiseStepCompiler();
	virtual ~LevelWiseStepCompiler();

	void Add (OperationItem* Node, int type);

	void CheckTreeLevel();

	int GetTreeLevel();

	void PrintParameters(int recursive);
};




//this is used to construct the operations
class OperationManager
{
public:
	int relations_no;
	int *relation_size;

	LevelWiseStepCompiler TableScanLevel;
	LevelWiseStepCompiler JoinLevel[MAXLEVEL];

	int ActualJoinLevel;
	LevelwiseStep **join_level;
	NewInMemoryJoin **imj_join_level;
	
	LevelwiseStep *aggregate_level;

	vector<OperationItem*> CurrentOperations;
	OperationItem *root;

	int *RunningOrder;

	char *schema_file;

	map<int,int> PipeIdtoIndex;
	vector<Pipe*> CurrentPipes;


	OperationManager(char *schema_file);
	virtual ~OperationManager();

	void BuiltQueryTree();

	void InitCurrentOperations();

	void SetupPipeIndex();

	void CheckTree();
	void CheckLevel();

	void RunAll();

	void PrepareRunningOrder();

	void BuildRelationSizes();

	void BuildJoinLevels();
	void BuildAggregateLevel();

	void BuildInMemoryJoinParameters();
	
	int GetNumGroupsToDisplay();
	int GetNumTuplesToStats();
	int GetQueueSize();
};


#endif


