#ifndef GROUP_IN_MEMORY_JOIN_WRAPPER_H
#define GROUP_IN_MEMORY_JOIN_WRAPPER_H

#include "ParseTreeWrapper.h"
#include "ExtendedHash.h"
#include "Comparison.h"
#include "Function.h"
#include "Record.h"
#include "Hashes.h"
#include "Function.h"
#include "Estimation.h"
#include "DEFS.h"

#include <map>
#include <vector>

using namespace std;



class NewInMemoryJoinWrapper
{
public:
	// the first param tells how many relations are taking place in the 
	// join.  The second parameter is a 2-D array of CNF expressions that
	// link up the input relations.  For example, myCNF[2][3] is the CNF
	// expression that is used to join the third and fourth input relations,
	// respecitvely.  mask tells which of the CNF expressions are actually
	// useful (since some of them might not apply to the query in question) 
	// if mask[i][j] is 1, then the corresponding CNF entry is used; if it
	// is zero, then the corresponding CNF entry is never used.  By convention,
	// myCNF[i][j] for i >= j is never used.  The myCNF array is consumed
	// by the constructor, so that after the call it should be 0.
	//The last two functions are used to generate CNFs and Function object
	//for the join query.
	NewInMemoryJoinWrapper(int numRels, int **mask, Schema **relSchemas, struct AndList *myCondition, struct FuncOperator *funcParseTree, double softBound, double hardBound, struct AttrGroupbyList *myGroupBy);

	~NewInMemoryJoinWrapper();
	
	//controller calls this to tell in memory join the constant 
	//relations, the first parameter tells the number of constant 
	//Relations, the second is an array of indices
	void InitialConstRel(int numConstRels, int *constRels);
	
	//controller insert records;
	SearchResults* ExternInsertRec(Record& rec, int whichRel, int timeStamp );
	
	//Set Statistics module that I need to use
	void SetStatModule (Estimation* fromMe);
	
	//insert a record with the relations, timestamp
	void InsertRec (Record& rec, int whichRel, int timeStamp);
	void InsertRec (Record& rec, int whichRel, int timeStamp, double ratio);
	
	//output Results, with indices and hash values, the last parameter
	//is the number of records in the output
	void outputResults(int**&hashes, double *& aggregates, int& numRecs);
	
	//obtain current sampling ratio for each relation
	void obtainSamplingRatio(double *&ratios);
	
	//This is used to prepare all necessary CNFs, Schemas
	//It should be called after the starting phase
	void PrepareParameters();
	
	//This hashes the constant records and add all startup tuples in the time order
	//furthermore if the first relation is a constant relation it adds all the tuple
	//from this relation to the join.
	//This must be called after PrepareParameters
	//for using in the start phase to determine the permutation order, this should not
	//be called
	void PreInsertRec();
	
	//set permutation order. called before PrepareParameters
	void SetPermOrder(int * orders);
	void GetPermOrder(int * orders);
	
	//added simply for debugging
	void Print();
	
	//add all non-const relation start tuples on sorted order based on timestamp;
	void AddStartTuples (Record &tuple, int timeStamp, int whichRel);
	
	//add all const relation tuples
	void AddConstTuples (Record* tuples,  int numTuples, int wichRel);
	
	//This chooses the permutaion order and clean all start up tuples after choosing the order
	SearchResults*  ChoosePermOrder(double *samplingRatio);
	SearchResults*  ChoosePermOrder();
	
	//pull out all the records;
	void PullOutAllJoinResults(Schema & outputSchema, vector<Record*>& myRecs);
	
	//Get Total number of Bytes occpuied by the join result
	double GetJoinSize();
	
	//Get Total number of records in the join result
	int GetResultTupleNumber();

//private:
	//information for groupby
	struct AttrGroupbyList *myGroupBy;
	int *groupAtts;
	int numGroupAtts;
	//The statistical module I will use
	Estimation* myStatModule;
	
	//totla number of relations
	int numRels;
	
	//mask matrix indicates the join relations
	int **mask;
	
	//all relation schemas
	Schema *relSchemas;
	
	//the whole condition andlist tree
	struct AndList *myCondition;
	
	//the function tree used to build function object that 
	//can be applied to the final join output tuple
	struct FuncOperator *funcParseTree;
	
	//the function built from the tree above
	Function *myFunc;
	
	//number of constant relations
	int numConstRels;
	
	//int array indicates which relations are constants
	int *constRels;
	
	double *constMask;
	//consttoPerm returns the permutation order of a constant relation
	//in the constRels, i.e if constRel[2] = 7 then, consttoPerm[7] =2
	map <int, int> consttoPerm;
	
	//the soft and hard bound of memory usage
	double softBound, hardBound;
	
	//we need some structure to store the constant relations
	struct ConstRelPair
	{
		int relIndex;
		RecordArray *myRecs;
		ExtendedHash * myExtendedHash;
		ConstRelPair();
		~ConstRelPair();
	};
	
	struct ConstRelPair *myConstRelRecs;
	
	//IndextoPerm gives index and returns permutation order
	//PermtoIndex gives is saved in the array PermutationOrder
	int *PermutationOrder;
	map <int, int> IndextoPerm;
	
	//this is used to store all necessary information for partial results
	struct PartialJoinResults
	{
		
		//record is inserted here	
		RecordArray *myRecs;
	
		//hash index is built here
		ExtendedHash *myExtendedHash;
	
		//the schema of the record stored in myRecs
		Schema *outputSchema;
	
		//the ordermaker used to hash records in myRecs
		OrderMaker *resultHashingOrderMaker;
	
		//the ordermaker used to hash input relations in current permutation
		OrderMaker *inputHashingOrderMaker;	
	
		//we need to first apply the compare (two records version, in hash class)
		//then decide whether need to join or not.
		//the result cnf
		CNF *resultCNF;
	
		//the corresponding record literal
		Record *resultLit;
	
		//a hash object to hash input records. built using the input ordermaker above
		CWHash *myInputHash;
	
		//A variable indicates whether the INPUT RELATION is a const relation or not
		bool isConst;
	
		//the dropping ratio for current relation
		double ratio;
		
		//The most recent timestamp and coinflip
		//if we see another record with the same time stamp then we use this coinflip instead of calling drand48()
		// they are initialized to be -1;
		int currentTimeStamp;
		double currentCoinFlip;
		
		// destructor
		~PartialJoinResults();
	};
	//all partial resPriorityQueue.o MMIndex.o SMOrder.oults
	struct PartialJoinResults *myPartialResults;
	
	//this hash object is defined for PURE comparison
	//Florin integrated the compare functions into the base hash object
	//which might be a bad idea, need to change later
	Hash myCompareHash;
	
	//the inserted tuples with timestamp and aggregate values
	vector <int *> * insertedTimeStamps;
	vector <double> * insertedAggregates;
	vector <Record*> *insertedGroups;
	
	//the deleted tuples with timestamp and aggregate values
	vector <int *> * deletedTimeStamps;
	vector <double> * deletedAggregates;
	vector <Record*> * deletedGroups;
	
	//here is used to store all the records for the starting phase
	//which should all be cleaned after the permutation order is determined
	RecordArray *myStartTuples;
	
	//after the start phase is finished, all thest tuples are added on order
	RecordArray *allStartTuples;
	
	vector <int> *relations; //used to store the relations for the allstarttuples
	
	inline double min(double a, double b);

	inline double max(double a, double b);
	
	//check the buffer exceeds the hardBound or not, if it exceeds the hardBound, determine
	//which relation to drop
	void CheckExceed();
	
	//dropping samples from a particular relation(in permutation order)
	//relIndex relation index, newRatio, the newRatio to keep
	void Dropping(int relIndex, double newRatio);

	void Erase();

	//for debugging
	int * numTuplesReceived;
};

#endif
