#include "PriorityQueue.h"

#include <string.h>
#include <stdio.h>
#include <iostream.h>



PriorityQueue :: PriorityQueue (char *run_name, int runs_no, OrderMaker *order, Hash *sort_func, Schema *schema, int buffer_size) {

	this->run_name = strdup(run_name);
	this->runs_no = runs_no;

	this->order = order;
	this->sort_func = sort_func;
	this->schema = schema;

	this->buffer_size = buffer_size;

	//STRIPE
//	this->run_file = new (std::nothrow) Run_Stripe_File();
	this->run_file = new (std::nothrow) Run_File();

	if (this->run_file == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	InitializeHeap();
}


PriorityQueue :: ~PriorityQueue() {
	run_file->Kill();
	delete run_file;
	run_file = NULL;

	free(run_name);
	run_name = NULL;

	sort_func = NULL;
	schema = NULL;

	for (int i = 0; i < nodes_no; i++) {
		delete [] external_heap[i].mem_run->buffer;
		delete external_heap[i].mem_run;
	}

	delete [] internal_heap;
	delete [] external_heap;

	nodes_no = 0;
}



void PriorityQueue :: InitializeHeap () {

	nodes_no = 1;
	while (nodes_no < runs_no)
		nodes_no = nodes_no << 1;

	if (nodes_no == 1)
		nodes_no = 2;

	// distribute the memory to each run
	int mem_run_size = buffer_size / runs_no;
	if (mem_run_size < 1) {
		cerr << "More runs than buffer can hold!\n";
		exit(0);
	}

	run_file->Open(1, run_name);

	// create heap nodes
	internal_heap = new (std::nothrow) InternalNode[nodes_no];
	if (internal_heap == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	external_heap = new (std::nothrow) ExternalNode[nodes_no];
	if (external_heap == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	// initialize the external nodes
	for (int i = 0; i < nodes_no; i++) {
		// create an inmemory run
		external_heap[i].mem_run = new (std::nothrow) InMemoryRun;
		if (external_heap[i].mem_run == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		external_heap[i].mem_run->buffer_size = mem_run_size;

		if (i < runs_no) {
			external_heap[i].run_no = i;
			external_heap[i].mem_run->buffer = new (std::nothrow) Record[mem_run_size];
			if (external_heap[i].mem_run->buffer == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			ReadDataIntoRun(external_heap[i].mem_run, external_heap[i].run_no);

			external_heap[i].mem_run->cur_index = 0;

			// check for empty run file
			if (external_heap[i].mem_run->cur_size > 0) {
				external_heap[i].rec = &external_heap[i].mem_run->buffer[external_heap[i].mem_run->cur_index];

				external_heap[i].mem_run->cur_index += 1;
			}
			else {
				external_heap[i].rec = NULL;
			}
		}
		
		else {
			external_heap[i].run_no = -1;

			external_heap[i].mem_run->buffer = NULL;

			external_heap[i].mem_run->cur_size = 0;
			external_heap[i].mem_run->cur_index = 0;

			external_heap[i].rec = NULL;
		}
	}

	int child_index;

	// initialzie the first internal level, from external level
	int cur_level = nodes_no >> 1;
	for (int i = cur_level; i < nodes_no; i++) {
		child_index = (i << 1) - nodes_no;
		UpdateFromExternal(i, child_index);
	}

	// initialze rest of the internal levels
	cur_level = cur_level >> 1;
	while (cur_level > 0) {
		for(int i = cur_level; i < (cur_level << 1); i++) {
			child_index = i << 1;
			UpdateFromInternal(i, child_index);
		}

		cur_level = cur_level >> 1;
	}
}



void PriorityQueue :: ReadDataIntoRun(InMemoryRun *run_mem, int run_no) {
	Record *temp_rec = new (std::nothrow) Record();
	if (temp_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int i;
	for (i = 0; i < run_mem->buffer_size; i++) {
		if (run_file->NextRec(temp_rec, run_no)) {

//			temp_rec->PrintToFile(schema, run_name);

			temp_rec->hash = sort_func->Compute_Hash(temp_rec, order);
			run_mem->buffer[i].Consume(temp_rec);
		}
		else {
			break;
		}
	}

	delete temp_rec;

	run_mem->cur_size = i;
}


void PriorityQueue :: UpdateFromExternal(int parent_index, int child_index) {
	internal_heap[parent_index].rec = external_heap[child_index].rec;
	internal_heap[parent_index].which_external = child_index;

	if (internal_heap[parent_index].rec == NULL) {
		internal_heap[parent_index].rec = external_heap[child_index + 1].rec;
		internal_heap[parent_index].which_external = child_index + 1;
	}
	else if ((external_heap[child_index + 1].rec != NULL) && (sort_func->Compare(external_heap[child_index+1].rec, internal_heap[parent_index].rec, order) < 0)) {
		internal_heap[parent_index].rec = external_heap[child_index + 1].rec;
		internal_heap[parent_index].which_external = child_index + 1;
	}
}


void PriorityQueue :: UpdateFromInternal(int parent_index, int child_index) {
	internal_heap[parent_index].rec = internal_heap[child_index].rec;
	internal_heap[parent_index].which_external = internal_heap[child_index].which_external;

	if (internal_heap[parent_index].rec == NULL) {
		internal_heap[parent_index].rec = internal_heap[child_index + 1].rec;
		internal_heap[parent_index].which_external = internal_heap[child_index+1].which_external;
	}
	else if ((internal_heap[child_index + 1].rec != NULL) && (sort_func->Compare(internal_heap[child_index+1].rec, internal_heap[parent_index].rec, order) < 0)) {
		internal_heap[parent_index].rec = internal_heap[child_index + 1].rec;
		internal_heap[parent_index].which_external = internal_heap[child_index+1].which_external;
	}
}



int PriorityQueue :: CompareToTop(Record *record) {
	int res;

	if (internal_heap[1].rec == NULL) {
		res = 1;
	}
	else {
		res = sort_func->Compare(record, internal_heap[1].rec, order);
	}

	return res;
}


int PriorityQueue :: GetNext(Record *record) {
	// if root is empty, runs are exhausted
	if (internal_heap[1].rec == NULL) {
		return 0;
	}

	record->Consume(internal_heap[1].rec);

	int which_external = internal_heap[1].which_external;
	if (external_heap[which_external].mem_run->cur_index < external_heap[which_external].mem_run->cur_size) {
		external_heap[which_external].rec = &external_heap[which_external].mem_run->buffer[external_heap[which_external].mem_run->cur_index];

		external_heap[which_external].mem_run->cur_index += 1;
		
	}
	else {
		ReadDataIntoRun(external_heap[which_external].mem_run, external_heap[which_external].run_no);

		external_heap[which_external].mem_run->cur_index = 0;

		// check if any records are read
		if (external_heap[which_external].mem_run->cur_size > 0) {
			external_heap[which_external].rec = &external_heap[which_external].mem_run->buffer[external_heap[which_external].mem_run->cur_index];

			external_heap[which_external].mem_run->cur_index += 1;
		}
		else {
			external_heap[which_external].rec =  NULL;
		}
	}

	// do the percolate up on the new record inserted
	// this should be done even if the above record is NULL
	
	// first update the bottom most internal level
	int child1, child2;
	int parent;
	
	child1 = which_external;
	if (child1 % 2 == 0) {
		child2 = child1 + 1;
		parent = (child1 + nodes_no) >> 1;
		UpdateFromExternal(parent, child1);
	}
	else {
		child2 = child1 - 1;
		parent = (child2 + nodes_no) >> 1;
		UpdateFromExternal(parent, child2);
	}

	// update rest of the nodes in the path from external node to root of the heap
	int cur_level = nodes_no >> 2;
	while (cur_level > 0) {
		child1 = parent;
		if (child1 % 2 == 0) {
			child2 = child1 + 1;
			parent = child1 >> 1;
			UpdateFromInternal(parent, child1);
		}
		else {
			child2 = child1 - 1;
			parent = child2 >> 1;
			UpdateFromInternal(parent, child2);
		}

		cur_level = cur_level >> 1;
	}

	return 1;
}

