#ifndef RECORD_H
#define RECORD_H

#include <stdio.h>

#include "Schema.h"
#include "Comparison.h"
#include "ParseTree.h"


// Basic record data structure. Data is actually stored in "bits" field. The layout of bits is as
// follows:
//	1) First sizeof(int) bytes: length of the record in bytes
//	2) Next sizeof(int) bytes: byte offset to the start of the first att
//	3) Byte offset to the start of the att in position numAtts. There are as many offsets as
//	the number of attributes in the schema of the record
//	4) Bits encoding the record's data

class Record {
private:
	//the binary content of the record or the actual data in the record
	char *bits;

public:
	//records are sorted based on random hash values for the approximations to work
	//to avoid the recomputation of the hash value corresponding to a record,
	//it is stored in hash
	//careful dealing with hash is required because this value is changing from one level
	//to another
	//even for the same level, different hash functions generate different hash values
	int hash;


	Record ();
	virtual ~Record();


	//gives access to the internal bits of a record; not encapsulation anymore
	//it is used only when neccessary, no abbuse
	//this is preferable to friend definitions
	char* GetBits ();

	//this returns the specified attribute from the record
	char* GetColumn(int which_column);

	//returns the size of the record, that is, the first integer in the bits
	int GetSize();

	//copy bits of length b_len into the current record
	void CopyBits(char *bits, int b_len);

	// suck the contents of the record fromMe into this; note that after
	// this call, fromMe will no longer have anything inside of it
	void Consume (Record *fromMe);
	// this version sucks the binary content of a record
	void Consume (char *&fromMe);

	// makes the data bits NULL
	void Nullify ();

	// make a copy of the record fromMe; note that this is far more 
	// expensive (requiring a bit-by-bit copy) than Consume, which is
	// only a pointer operation
	void Copy (Record *copyMe);

	// reads the next record from a pointer to a text file; also requires
	// that the schema be given; returns a 0 if there is no data left or
	// if there is an error and returns a 1 otherwise
	int SuckNextRecord (Schema *mySchema, FILE *textFile);


	// this projects away attributes from the record
	// only the attributes specified in attsToKeep (the indices) are kept
	// numAttsToKeep gives the number of attributes to be kept
	// numAttsNow specifies the current number of attributes present in the record
	void Project (int *attsToKeep, int numAttsToKeep, int numAttsNow);

	//left and right records are merged into the current record object
	//numAttsLeft and numAttsRight specify the number of attributes to be kept
	//attsToKeep gives the indices of the attributes to be kept
	//startOfRight gives the index in attsToKeep where the attributes from right start
	//notice that while merging, projection is also applied
	void MergeRecords (Record *left, Record *right, int numAttsLeft, int numAttsRight, int *attsToKeep, int numAttsToKeep, int startOfRight);


	//left and right records are merged into the current record object
	//numAttsLeft and numAttsRight specify the number of attributes in Left and Right
	void AppendRecords (Record *left, Record *right, int numAttsLeft, int numAttsRight);



	// record Literal is extracted from parseTree into the current record object
	// the schema of the Literal record is extracted into mySchema
	// cnf is built from parseTree based on the left and right schemas as a secondary effect
	// this method has a problem because of its dual effects
	void GrowFromParseTree (struct AndList *parseTree, Schema *leftSchema, Schema *rightSchema, CNF &cnf, Schema *&mySchema);

	// version of the same function, except that it is used in the case of
	// a relational selection over a single relation so only one schema is used
	// the same problem of secondary effects persists in this case
	void GrowFromParseTree (struct AndList *parseTree, Schema *mySchema, CNF &cnf, Schema *&outSchema);


	// prints the contents of the record to the screen; this requires
	// that the schema also be given so that the record can be interpreted
	void Print (Schema *mySchema);
	void Print (Schema *mySchema, char *&output);
	//the same as Print, but the printing is done into the file "file_name"
	void PrintToFile (Schema *mySchema, char *file_name);
};



class SearchResults {
public:
	// this tells us the number of result tuples
	int numRes;
	// these tell us all of the result tuples that came
	// from the last insertion; the first is a pointer to
	// a number of arrays; each entry in the array corresponds
	// to a result tuple, and tells us the timestamp for each
	// input tuple making up the result tuple
	int **resTuples;
	// this tells us the function value associated with each
	// result tuple
	double *funcRes;
	//the grouping attributes for each discovered tuple
	Record *resGroups;
	
	
	int numDel;
	// this is similar to the above, but it tells us any tuples
	// that have been DELETED
	int **delTuples;
	// this tells us the function value associated with each
	// deleted tuple
	double *funcDel;
	//the grouping attributes for each dropped tuple
	Record *delGroups;
	
	// this tells us the current sampling fraction for each
	// of the input relations
	double *ratios;
	
	//this tells which relation is a constant relation
	//0 for nont-constant and 1 for constant
	// numConst contain the size of the array;
	// it is actually the number of relaitons, and the double array ratios
	// has the same size
	int *constRels;
	int numConst;
	
	// constructor and destructor
	SearchResults ();
	~SearchResults ();
	
	//for debugging
	void Print();
};


#endif
