#ifndef RELOP_LEVEL_H
#define RELOP_LEVEL_H

#include "Schema.h"
#include "File.h"
#include "Pipe.h"
#include "Comparison.h"
#include "Hashes.h"
#include "TwoWayList.h"
#include "SMOrder.h"
#include "Record.h"
#include "Function.h"
#include "ParseTree.h"
#include "ParseTreeWrapper.h"
#include "MMIndex.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include<vector>


struct ThreadTableScanParam {
	File_Interface *f;
	Pipe *out_pipe;
};




class RelationalOpLevel {

public:
	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size) = 0;

	// This tells the levelwise step that it is time to start up; this function
	// should be called after all of the operations have been added to the
	// step and before GetNextBatch is ever called
	virtual void StartUp () = 0;

	// This asks the i^th output pipe what percentage of its records still
	// remain to be processed
	virtual double HowMuchIsLeft () = 0;

	virtual double GetTuplesNo();
	virtual double GetAggregate();
	virtual Schema* GetSchema();
	virtual Function* GetFunction();
	virtual void GetJoinAtts(int *&join_atts, int &join_atts_no);
	virtual void GetGroupAtts(int *&group_atts, int &group_atts_no);

	virtual void GetGroups(vector<Record*> &desc, vector<double> &aggregate, vector<double> &tuples, int &groups_no);


	virtual ~RelationalOpLevel();
};


class TableScanLevel : public RelationalOpLevel {

public:
	//this constructor is needed only for inheritance purposes
	TableScanLevel();
	TableScanLevel(const char *file_location, Pipe *out_pipe, CNF *sel_op, Record *literal,
		       Schema *schema, int *atts, int atts_no, int tuples_no);
	virtual ~TableScanLevel();

	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size);

	virtual void StartUp ();

	virtual double HowMuchIsLeft ();

protected:
	pthread_t thr;
	pthread_mutex_t start_mutex;

	char *file_location;

	File_Interface *f;

	Pipe *pipe;
	CNF *op;
	Record *lit;
	Schema *schema;

	int *atts;
	int atts_no;

	int tuples_no;
	int tuples_proc;

	Hash *exact_hash;

	Hash *hash_time;

	const gsl_rng_type *T;
	gsl_rng *r;

	double last_time_stamp;
	double last_tuple_id;


	int GenerateTimeStamp(int tuple_id);

private:
	static void *DoStartUp(void *arg);
};




class TableScanLevelC : public TableScanLevel {

public:
	TableScanLevelC(const char *file_location, Pipe *out_pipe, CNF *sel_op, Record *literal,
		       Schema *schema, int *atts, int atts_no, int tuples_no);
	virtual ~TableScanLevelC();

	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size);

	virtual void StartUp ();

protected:
	//files and pipes from the input column files
	File_Interface **fc;
	Pipe **f_pipe;

	//thread parameters
	ThreadTableScanParam **thr_params;

	//threads, one for each of the columns in the output schema
	pthread_t *myThreads;

	//attributes that need to be read
	//they are either in the output schema or there are predicates on them
	int *read_atts;
	int read_atts_no;

private:
	static void *DoStartUp(void *arg);
};






class TableScanPreJoinLevel : public RelationalOpLevel {

public:
	TableScanPreJoinLevel(char **file_location, Pipe *out_pipe, Schema **schema, int **proj_atts, int *proj_atts_no, int *proj_atts_out, int proj_atts_out_no, AndList *conditions, int *tuples_no, int relations_no);
	virtual ~TableScanPreJoinLevel();

	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size);

	virtual void StartUp ();

	virtual double HowMuchIsLeft ();

protected:
	pthread_t thr;
	pthread_mutex_t start_mutex;

	int small_relations_no;

	char *big_file_location;
	char **file_location;

	File_Interface *big_file;
	File_Interface **f;

	Pipe *pipe;

	Schema *big_file_schema;
	Schema **schema;

	int *proj_atts_no;
	int **proj_atts;

	int small_atts_no, total_atts_no;

	int proj_atts_out_no;
	int *proj_atts_out;

	AndListWrapper *conditions;
	OrderMaker small_order, big_order;

	Record *recLiteral;
	CNF *cnf;

	MMIndex *hash_table;

	int big_file_tuples_no;
	int tuples_proc;

	int *tuples_no;

	Hash *exact_hash;

	const gsl_rng_type *T;
	gsl_rng *r;

	double last_time_stamp;
	double last_tuple_id;


	int GenerateTimeStamp(int tuple_id);

private:
	static void *DoStartUp(void *arg);
};





class SortJoinLevel : public RelationalOpLevel {

public:
	SortJoinLevel(Pipe *in_pipe_left, int pipe_left_index, Pipe *in_pipe_right, int pipe_right_index, Pipe *out_pipe, CNF *sel_op, Record *literal, Schema *schema_left, Schema *schema_right, int *atts, int atts_no, int buffer_size);
	virtual ~SortJoinLevel();

	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size);

	virtual void StartUp ();

	virtual double HowMuchIsLeft ();

private:
	int domain_proc;
	int buffer_size;
	
	OrderMaker *om_left, *om_right;
	Order *order_left, *order_right;
	
	Pipe *in_pipe_left, *in_pipe_right, *out_pipe;
	
	CNF *op;
	CNF *extra_cond;

	Record *lit;
	Schema *schema_left, *schema_right;
	
	int *atts;
	int atts_no;
	int right_start;

	Hash *sort_func;
	Hash *exact_hash;

	int which_index;

	int is_done;

	TwoWayList<Record> *cur_left, *cur_right;


	pthread_t thread1, thread2, thread3;
	pthread_mutex_t start_mutex;



	// the actual writeout method.
	static void *DoSort(void *arg);

	static void *ReadRight(void *arg);
	static void *ReadLeft(void *arg);

};




class RehashLevel : public RelationalOpLevel {

public:
	RehashLevel(Pipe *in_pipe, int pipe_index, Pipe *out_pipe, Schema *schema, int buffer_size);
	virtual ~RehashLevel();

	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size);

	virtual void StartUp ();

	virtual double HowMuchIsLeft ();

private:
	int domain_proc;
	int buffer_size;
	
	OrderMaker *om;
	Order *order;
	
	Pipe *in_pipe, *out_pipe;
	
	Schema *schema;
	
	Hash *sort_func;

	int is_done;

	TwoWayList<Record> *cur_rec;


	pthread_t thread1;
	pthread_mutex_t start_mutex;



	// the actual writeout method.
	static void *DoSort(void *arg);

};





class WriteOutLevel : public RelationalOpLevel {

public:
	WriteOutLevel(Pipe *pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no);
	virtual ~WriteOutLevel();

	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size);

	virtual void StartUp ();

	virtual double HowMuchIsLeft ();

	virtual double GetTuplesNo();
	virtual double GetAggregate();
	virtual Schema* GetSchema();
	virtual Function* GetFunction();
	virtual void GetJoinAtts(int *&join_atts, int &join_atts_no);


private:
	pthread_t thr;
	pthread_mutex_t start_mutex;

	Pipe *pipe;
	Schema *schema;
	Function *func;

	double tuples_no;
	double tuples_proc;

	File_Interface *f;
	char file_name[100];

	double aggregate;

	int *join_atts;
	int join_atts_no;


	static void *DoStartUp(void *args);
};




class GroupAggregateLevel : public RelationalOpLevel {

public:
	GroupAggregateLevel(Pipe *pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no, int *group_atts, int group_atts_no, int buffer_size);
	virtual ~GroupAggregateLevel();

	virtual int GetNextBatch (double p, Record *output, int max_output_size, int &output_size);

	virtual void StartUp ();

	virtual double HowMuchIsLeft ();

	virtual double GetTuplesNo();
	virtual Schema* GetSchema();
	virtual Function* GetFunction();
	virtual void GetJoinAtts(int *&join_atts, int &join_atts_no);
	virtual void GetGroupAtts(int *&group_atts, int &group_atts_no);

	virtual void GetGroups(vector<Record*> &desc, vector<double> &aggregate, vector<double> &tuples, int &groups_no);


private:
	pthread_t thr;
	pthread_mutex_t start_mutex;

	Pipe *pipe;
	Schema *schema;
	Function *func;

	double tuples_no;
	double tuples_proc;

	int *join_atts;
	int join_atts_no;

	int *group_atts;
	int group_atts_no;

	Schema *store_schema;
	OrderMaker *store_order;

	OrderMaker *group_order;
	Order *order;

	vector<Record*> groups;
	vector<double> group_aggs;
	vector<double> group_tuples;

	int buffer_size;

	Hash *sort_func;
	TwoWayList<Record> *cur_rec;


	static void *DoStartUp(void *args);
};




//global thread function for C-STORE table scan
void *TableScan_Thread (void *param);


#endif


