#include "SamplingEstimate.h"
#include "StatStruct.h"

#include <iostream>
#include <vector>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_pow_int.h>


using namespace std;


SamplingData::SamplingData()
{
	relations_no = 0;
}

SamplingData::SamplingData(int relations_no)
{
	this->relations_no = relations_no;
}

SamplingData::~SamplingData()
{
	relations_no = 0;
}





SimpleSamplingData::SimpleSamplingData(int relations_no, double *p)
: SamplingData(relations_no)
{
	this->type = Bernoulli;

	e = new (std::nothrow) double[relations_no];
	if (e == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	a_e2 = new (std::nothrow) double[relations_no];
	if (a_e2 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	b_e2 = new (std::nothrow) double[relations_no];
	if (b_e2 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < relations_no; i++)
	{
		e[i] = p[i];
		a_e2[i] = 1.0;
		b_e2[i] = (1.0 / p[i] - 1.0);
	}
}


SimpleSamplingData::SimpleSamplingData(int relations_no, int *rel_size, int *samp_size)
: SamplingData(relations_no)
{
	this->type = WOR;

	e = new (std::nothrow) double[relations_no];
	if (e == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	a_e2 = new (std::nothrow) double[relations_no];
	if (a_e2 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	b_e2 = new (std::nothrow) double[relations_no];
	if (b_e2 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	
	for (int i = 0; i < relations_no; i++)
	{
		e[i] = (double)samp_size[i] / (double)rel_size[i];
		a_e2[i] = (double)rel_size[i] * (1.0 - 1.0 / (double)samp_size[i]) / ((double)rel_size[i] - 1.0);
		b_e2[i] = (double)rel_size[i] * ((double)rel_size[i] / (double)samp_size[i] - 1.0) / ((double)rel_size[i] - 1.0);
	}
}


SimpleSamplingData::SimpleSamplingData(const SimpleSamplingData &sd)
{
	this->type = sd.type;

	this->relations_no = sd.relations_no;

	e = new (std::nothrow) double[relations_no];
	if (e == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	memcpy(e, sd.e, relations_no * sizeof(double));

	a_e2 = new (std::nothrow) double[relations_no];
	if (a_e2 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	memcpy(a_e2, sd.a_e2, relations_no * sizeof(double));

	b_e2 = new (std::nothrow) double[relations_no];
	if (b_e2 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	memcpy(b_e2, sd.b_e2, relations_no * sizeof(double));
}


SimpleSamplingData::~SimpleSamplingData()
{
	delete [] e;
	e = NULL;

	delete [] a_e2;
	a_e2 = NULL;

	delete [] b_e2;
	b_e2 = NULL;

	relations_no = 0;
}


double SimpleSamplingData::ComputeCCoeff(int S)
{
	return 0.0;
}


double SimpleSamplingData::ComputeCCoeff(int S, int T)
{
	return 0.0;
}


void SimpleSamplingData::Print()
{
	cout << "E:\n";
	for (int i = 0; i < relations_no; i++)
	{
		printf("%10.8f\t", e[i]);
	}
	cout << "\n";

	cout << "A-E2:\n";
	for (int i = 0; i < relations_no; i++)
	{
		printf("%10.8f\t", a_e2[i]);
	}
	cout << "\n";

	cout << "B-E2:\n";
	for (int i = 0; i < relations_no; i++)
	{
		printf("%10.8f\t", b_e2[i]);
	}
	cout << "\n";
}







GeneralizedSamplingData::GeneralizedSamplingData(int relations_no, int constant_relations, int constant_mask, double threshold, double *bernoulli_p)
: SamplingData(relations_no)
{
	this->constant_relations = constant_relations;
	this->constant_mask = constant_mask;

	this->threshold = threshold;

	this->bernoulli_p = new (std::nothrow) double[this->relations_no];
	if (this->bernoulli_p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	memcpy(this->bernoulli_p, bernoulli_p, relations_no * sizeof(double));


//	cout << "Threshold = " << threshold << "\n";
//	cout << "Constant Mask = " << this->constant_mask << "\n";


	//compute the scaling factor for the expectation
	double fact = gsl_sf_fact(this->relations_no - this->constant_relations);
	double thr = gsl_pow_int(this->threshold, this->relations_no - this->constant_relations);

	this->a = thr / fact;

	for (int i = 0; i < this->relations_no; i++)
	{
		this->a *= this->bernoulli_p[i];
	}


	//compute the variance coefficients
	this->b_t = new (std::nothrow) double[1 << this->relations_no];
	if (this->b_t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		//compute the b_t coefficient for each possible combination of relations
		b_t[i] = 1.0;

		//first, deal with the Bernoulli probabilities
		for (int j = 0; j < this->relations_no; j++)
		{
			//S contains j
			if ((i & (1 << j)) != 0)
			{
				b_t[i] *= this->bernoulli_p[j];
			}
			//S does not contain j
			else
			{
				b_t[i] *= (this->bernoulli_p[j] * this->bernoulli_p[j]);
			}
		}

//		printf("b[%d] = %f\t", i, b_t[i]);

		//second, deal with the threashold probabilities
		for (int j = 0; j < this->relations_no; j++)
		{

			// skip over constant relations
			if ((this->constant_mask & (1 << j)) != 0)
				continue;

			if ((i & (1 << j)) != 0)
			{
				b_t[i] *= this->threshold;
			}
			else
			{
				b_t[i] *= (this->threshold * this->threshold);
			}
		}

//		printf("%f\t", b_t[i]);

		//third, deal with special factorial function
		double f_factorial = Compute_Double_Prob(i, this->relations_no, this->constant_mask);

//		printf("%10.8f\n", f_factorial);

		b_t[i] *= f_factorial;

//		printf("%f\n", b_t[i]);
	}
}


GeneralizedSamplingData::GeneralizedSamplingData(int relations_no, int constant_relations, int constant_mask, double threshold, vector<double> *bernoulli_p)
: SamplingData(relations_no)
{
	this->constant_relations = constant_relations;
	this->constant_mask = constant_mask;

	this->threshold = threshold;

	this->bernoulli_p = new (std::nothrow) double[this->relations_no];
	if (this->bernoulli_p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	for (int i = 0; i < relations_no; i++)
	{
		this->bernoulli_p[i] = (*bernoulli_p)[i];
	}


//	cout << "Threshold = " << threshold << "\n";
//	cout << "Constant Mask = " << this->constant_mask << "\n";


	//compute the scaling factor for the expectation
	double fact = gsl_sf_fact(this->relations_no - this->constant_relations);
	double thr = gsl_pow_int(this->threshold, this->relations_no - this->constant_relations);

	this->a = thr / fact;

	for (int i = 0; i < this->relations_no; i++)
	{
		this->a *= this->bernoulli_p[i];
	}


	//compute the variance coefficients
	this->b_t = new (std::nothrow) double[1 << this->relations_no];
	if (this->b_t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		//compute the b_t coefficient for each possible combination of relations
		b_t[i] = 1.0;

		//first, deal with the Bernoulli probabilities
		for (int j = 0; j < this->relations_no; j++)
		{
			//S contains j
			if ((i & (1 << j)) != 0)
			{
				b_t[i] *= this->bernoulli_p[j];
			}
			//S does not contain j
			else
			{
				b_t[i] *= (this->bernoulli_p[j] * this->bernoulli_p[j]);
			}
		}

//		printf("b[%d] = %f\t", i, b_t[i]);

		//second, deal with the threashold probabilities
		for (int j = 0; j < this->relations_no; j++)
		{

			// skip over constant relations
			if ((this->constant_mask & (1 << j)) != 0)
				continue;

			if ((i & (1 << j)) != 0)
			{
				b_t[i] *= this->threshold;
			}
			else
			{
				b_t[i] *= (this->threshold * this->threshold);
			}
		}

//		printf("%f\t", b_t[i]);

		//third, deal with special factorial function
		double f_factorial = Compute_Double_Prob(i, this->relations_no, this->constant_mask);
		b_t[i] *= f_factorial;

//		printf("%f\n", b_t[i]);
	}
}


GeneralizedSamplingData::~GeneralizedSamplingData()
{
	delete [] bernoulli_p;
	bernoulli_p = NULL;

	delete b_t;
	b_t = NULL;

	relations_no = 0;
	constant_relations = 0;
}



double GeneralizedSamplingData::ComputeCCoeff(int S)
{
	double res = 0.0;

//	printf("C(%d)\n", S);

	int T = 0;
	while (T <= S)
	{
		//skip over T not in POWER(S)
		if ((T & ~S) != 0)
		{
			T += 1;
			continue;
		}
		
		int ip = S & ~T;

		double coeff = 1.0;
		for (int i = 0; i < relations_no; i++)
		{
			// now treat bit in scan as in the big variance since these are variance terms
			if ((ip & (1 << i)) != 0)
				coeff *= -1.0;
		}

		res += coeff * b_t[T];

//		printf("%d : %f %f\n", T, coeff, b_t[T]);

		T += 1;
	}

	return res;
}


double GeneralizedSamplingData::ComputeCCoeff(int S, int T)
{
	double res = 0.0;

//	printf("C(%d,%d)\n", S, T);

	int U = 0;
	while (U <= T)
	{
		//skip over U not in POWER(T)
		if ((U & ~T) != 0)
		{
			U += 1;
			continue;
		}

		//ip = T - U
		int ip = T & ~U;

		double coeff = 1.0;
		for (int i = 0; i < relations_no; i++)
		{
			// now treat bit in scan as in the big variance since these are variance terms
			if ((ip & (1 << i)) != 0)
				coeff *= -1.0;
		}

		int s_union_u = U | S;
		res += coeff * b_t[s_union_u];

//		printf("%d : %f %f\n", ip, coeff, b_t[s_union_u]);

		U += 1;
	}

	return res;
}


void GeneralizedSamplingData::Print()
{
	cout << "Threshold : ";
	printf("%10.8f\n", threshold);

	cout << "Bernoulli probabilities:\n";
	for (int i = 0; i < relations_no; i++)
	{
		printf("%10.8f\t", bernoulli_p[i]);
	}
	cout << "\n";

	cout << "A : ";
	printf("%30.10f\n", a);

	cout << "B :\n";
	for (int i = 0; i < (1 << relations_no); i++)
	{
		printf("%30.10f\n", b_t[i]);
	}
}




GeneralizedSamplingDataCov::GeneralizedSamplingDataCov(int relations_no, int constant_relations, int constant_mask, double threshold1, double threshold2, vector<double> *bernoulli_p1, vector<double> *bernoulli_p2)
: SamplingData(relations_no)
{
	this->constant_relations = constant_relations;
	this->constant_mask = constant_mask;

	this->threshold1 = threshold1;
	this->threshold2 = threshold2;

	this->bernoulli_p1 = bernoulli_p1;
	this->bernoulli_p2 = bernoulli_p2;


//	cout << "Threshold = " << threshold << "\n";
//	cout << "Constant Mask = " << this->constant_mask << "\n";


	//compute the scaling factor for the expectation
	double fact = gsl_sf_fact(this->relations_no - this->constant_relations);
	double thr = gsl_pow_int(this->threshold1, this->relations_no - this->constant_relations);

	this->a1 = thr / fact;

	for (int i = 0; i < this->relations_no; i++)
	{
		this->a1 *= (*this->bernoulli_p1)[i];
	}


	thr = gsl_pow_int(this->threshold2, this->relations_no - this->constant_relations);

	this->a2 = thr / fact;

	for (int i = 0; i < this->relations_no; i++)
	{
		this->a2 *= (*this->bernoulli_p2)[i];
	}



	//compute the variance coefficients
	this->b_t = new (std::nothrow) double[1 << this->relations_no];
	if (this->b_t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		//compute the b_t coefficient for each possible combination of relations
		b_t[i] = 0.0;

		double b_tt = 1.0;

		//first, deal with the Bernoulli probabilities
		for (int j = 0; j < this->relations_no; j++)
		{
			//S contains j
			if ((i & (1 << j)) != 0)
			{
				b_tt *= (*this->bernoulli_p2)[j];
			}
			//S does not contain j
			else
			{
				b_tt *= ((*this->bernoulli_p1)[j] * (*this->bernoulli_p2)[j]);
			}
		}

		int i_mod = 0;
		int kp = 0;
		for (int j = 0; j < this->relations_no; j++)
		{
			if ((this->constant_mask & (1 << j)) == 0)
			{
				if ((i & (1 << j)) != 0)
				{
					i_mod = i_mod | (1 << kp);
				}
	
				kp += 1;
			}
		}

		//find the position of the most significant 1 bit in i
		int n_chain = 0;
		for (int j = kp - 1; j >= 0; j--)
		{
			//S contains j
			if ((i_mod & (1 << j)) != 0)
			{
				break;
			}
			else
			{
				n_chain += 1;
			}
		}
		n_chain = kp - n_chain;

		//second, deal with the threashold probabilities
		for (int j = 0; j < n_chain; j++)
		{
			if ((i_mod & (1 << j)) != 0)
			{
				b_tt *= this->threshold1;
			}
			else
			{
				b_tt *= (this->threshold1 * this->threshold1);
			}
		}


		int n_real = kp;
		int n_chain_real = n_chain;
		int n_not_chain_real = kp - n_chain;


		for (int j = 0; j < n_not_chain_real + 1; j++)
		{
			double tt = b_tt;

			//second, deal with the threashold probabilities
			for (int k = 0; k < j; k++)
			{
				tt *= (this->threshold1 * this->threshold1);
			}

			//second, deal with the threashold probabilities
			for (int k = j; k < n_not_chain_real; k++)
			{
				tt *= (this->threshold1 * (this->threshold2 - this->threshold1));
			}

			//third, deal with special factorial function
			double f_factorial = Compute_Double_Prob_Cov_No_Const(i_mod, n_real, n_chain_real, j);
			b_t[i] += f_factorial * tt;

/*
			printf("I NEW = %5d OLD = %5d\n", i_mod, i);
			printf("N NEW = %5d OLD = %5d\n", n_real, this->relations_no);
			printf("N CHAIN = %5d\n", n_chain_real);
			printf("LAST IN = %5d\n", j);
			printf("N NOT CHAIN = %5d\n", n_not_chain_real);
			printf("MASK = %5d\n", this->constant_mask);

			printf("NEW = %10.8f\n", f_factorial);
			printf("OLD = %10.8f\n", f_factorial_old);
*/
		}
	}
}


/*
GeneralizedSamplingDataCov::GeneralizedSamplingDataCov(int relations_no, int constant_relations, int constant_mask, double threshold1, double threshold2, vector<double> *bernoulli_p1, vector<double> *bernoulli_p2)
: SamplingData(relations_no)
{
	this->constant_relations = constant_relations;
	this->constant_mask = constant_mask;

	this->threshold1 = threshold1;
	this->threshold2 = threshold2;

	this->bernoulli_p1 = bernoulli_p1;
	this->bernoulli_p2 = bernoulli_p2;


//	cout << "Threshold = " << threshold << "\n";
//	cout << "Constant Mask = " << this->constant_mask << "\n";


	//compute the scaling factor for the expectation
	double fact = gsl_sf_fact(this->relations_no - this->constant_relations);
	double thr = gsl_pow_int(this->threshold1, this->relations_no - this->constant_relations);

	this->a1 = thr / fact;

	for (int i = 0; i < this->relations_no; i++)
	{
		this->a1 *= (*this->bernoulli_p1)[i];
	}


	thr = gsl_pow_int(this->threshold2, this->relations_no - this->constant_relations);

	this->a2 = thr / fact;

	for (int i = 0; i < this->relations_no; i++)
	{
		this->a2 *= (*this->bernoulli_p2)[i];
	}



	//compute the variance coefficients
	this->b_t = new (std::nothrow) double[1 << this->relations_no];
	if (this->b_t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		//compute the b_t coefficient for each possible combination of relations
		b_t[i] = 0.0;

		double b_tt = 1.0;

		//find the position of the most significant 1 bit in i
		int n_chain = 0;
		for (int j = this->relations_no - 1; j >= 0; j--)
		{
			//S contains j
			if ((i & (1 << j)) != 0)
			{
				break;
			}
			else
			{
				n_chain += 1;
			}
		}
		n_chain = this->relations_no - n_chain;

		//first, deal with the Bernoulli probabilities
		for (int j = 0; j < this->relations_no; j++)
		{
			//S contains j
			if ((i & (1 << j)) != 0)
			{
				b_tt *= (*this->bernoulli_p2)[j];
			}
			//S does not contain j
			else
			{
				b_tt *= ((*this->bernoulli_p1)[j] * (*this->bernoulli_p2)[j]);
			}
		}

		//second, deal with the threashold probabilities
		for (int j = 0; j < n_chain; j++)
		{
			// skip over constant relations
			if ((this->constant_mask & (1 << j)) != 0)
				continue;

			if ((i & (1 << j)) != 0)
			{
				b_tt *= this->threshold1;
			}
			else
			{
				b_tt *= (this->threshold1 * this->threshold1);
			}
		}

		int all_const = 1;
		int old_only_const_in = 0;
		int old_only_const_out = 0;

		for (int j = 0; j < this->relations_no - n_chain + 1; j++)
		{
			double tt = b_tt;

			int only_const_in = 0;

			//second, deal with the threashold probabilities
			for (int k = n_chain; k < n_chain + j; k++)
			{
				// skip over constant relations
				if ((this->constant_mask & (1 << k)) != 0)
					continue;
	
				tt *= (this->threshold1 * this->threshold1);

				only_const_in += 1;
			}

			int only_const_out = 0;

			//second, deal with the threashold probabilities
			for (int k = n_chain + j; k < this->relations_no; k++)
			{
				// skip over constant relations
				if ((this->constant_mask & (1 << k)) != 0)
					continue;
	
				tt *= (this->threshold1 * (this->threshold2 - this->threshold1));

				only_const_out += 1;
			}

			if (((only_const_out > 0) && (j == 0)) || ((only_const_in > 0) && (only_const_out > 0) && (old_only_const_in + 1 == only_const_in) && (old_only_const_out == only_const_out + 1)) || ((only_const_in > 0) && (j == this->relations_no - n_chain) && (old_only_const_in + 1 == only_const_in)))
			{
				//third, deal with special factorial function
				double f_factorial = Compute_Double_Prob_Cov(i, this->relations_no, this->constant_mask, n_chain, j);
				b_t[i] += f_factorial * tt;

//				printf("[%d] : %10.8f %10.8f %10.8f\n", i, f_factorial, tt, b_t[i]);

				all_const = 0;
			}

			old_only_const_in = only_const_in;
			old_only_const_out = only_const_out;
		}

		if (all_const == 1)
		{
			double f_factorial = Compute_Double_Prob_Cov(i, this->relations_no, this->constant_mask, n_chain, 0);
			b_t[i] = f_factorial * b_tt;
		}

//		printf("%f\n", b_t[i]);
	}
}
*/



GeneralizedSamplingDataCov::~GeneralizedSamplingDataCov()
{
	delete b_t;
	b_t = NULL;

	relations_no = 0;
	constant_relations = 0;
}



double GeneralizedSamplingDataCov::ComputeCCoeff(int S)
{
	double res = 0.0;

//	printf("C(%d)\n", S);

	int T = 0;
	while (T <= S)
	{
		//skip over T not in POWER(S)
		if ((T & ~S) != 0)
		{
			T += 1;
			continue;
		}
		
		int ip = S & ~T;

		double coeff = 1.0;
		for (int i = 0; i < relations_no; i++)
		{
			// now treat bit in scan as in the big variance since these are variance terms
			if ((ip & (1 << i)) != 0)
				coeff *= -1.0;
		}

		res += coeff * b_t[T];

//		printf("%d : %f %f\n", T, coeff, b_t[T]);

		T += 1;
	}

	return res;
}


double GeneralizedSamplingDataCov::ComputeCCoeff(int S, int T)
{
	double res = 0.0;

//	printf("C(%d,%d)\n", S, T);

	int U = 0;
	while (U <= T)
	{
		//skip over U not in POWER(T)
		if ((U & ~T) != 0)
		{
			U += 1;
			continue;
		}

		//ip = T - U
		int ip = T & ~U;

		double coeff = 1.0;
		for (int i = 0; i < relations_no; i++)
		{
			// now treat bit in scan as in the big variance since these are variance terms
			if ((ip & (1 << i)) != 0)
				coeff *= -1.0;
		}

		int s_union_u = U | S;
		res += coeff * b_t[s_union_u];

//		printf("%d : %f %f\n", ip, coeff, b_t[s_union_u]);

		U += 1;
	}

	return res;
}


void GeneralizedSamplingDataCov::Print()
{
	cout << "Threshold1 : ";
	printf("%10.8f\n", threshold1);

	cout << "Bernoulli probabilities1:\n";
	for (int i = 0; i < relations_no; i++)
	{
		printf("%10.8f\t", (*bernoulli_p1)[i]);
	}
	cout << "\n";

	cout << "Threshold2 : ";
	printf("%10.8f\n", threshold2);

	cout << "Bernoulli probabilities2:\n";
	for (int i = 0; i < relations_no; i++)
	{
		printf("%10.8f\t", (*bernoulli_p2)[i]);
	}
	cout << "\n";

	cout << "A1 : ";
	printf("%30.10f\n", a1);

	cout << "A2 : ";
	printf("%30.10f\n", a2);

	cout << "B :\n";
	for (int i = 0; i < (1 << relations_no); i++)
	{
		printf("%30.10f\n", b_t[i]);
	}
}





GeneralizedSamplingDataNoDiscard::GeneralizedSamplingDataNoDiscard(int relations_no, int constant_relations, int constant_mask, vector<double> *threshold, vector<double> **bernoulli_p)
: SamplingData(relations_no)
{
	this->constant_relations = constant_relations;
	this->constant_mask = constant_mask;

	this->threshold = threshold;
	this->bernoulli_p = bernoulli_p;

	//compute the scaling factor for the expectation
	double fact = gsl_sf_fact(this->relations_no - this->constant_relations);
	double thr = ComputeTCoeff();

	this->a = thr / fact;

	//compute the variance coefficients
	this->b_t = new (std::nothrow) double[1 << this->relations_no];
	if (this->b_t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		//compute the b_t coefficient for each possible combination of relations
		double thr_var = ComputeTVarCoeff(i);
		b_t[i] = thr_var;

//		printf("%f\t", b_t[i]);

		//third, deal with special factorial function
		double f_factorial = Compute_Double_Prob(i, this->relations_no, this->constant_mask);
		b_t[i] *= f_factorial;

//		printf("%f\n", b_t[i]);
	}
}


GeneralizedSamplingDataNoDiscard::~GeneralizedSamplingDataNoDiscard()
{
	delete b_t;
	b_t = NULL;

	relations_no = 0;
	constant_relations = 0;
}



double GeneralizedSamplingDataNoDiscard::ComputeCCoeff(int S)
{
	double res = 0.0;

//	printf("C(%d)\n", S);

	int T = 0;
	while (T <= S)
	{
		//skip over T not in POWER(S)
		if ((T & ~S) != 0)
		{
			T += 1;
			continue;
		}
		
		int ip = S & ~T;

		double coeff = 1.0;
		for (int i = 0; i < relations_no; i++)
		{
			// now treat bit in scan as in the big variance since these are variance terms
			if ((ip & (1 << i)) != 0)
				coeff *= -1.0;
		}

		res += coeff * b_t[T];

//		printf("%d : %f %f\n", T, coeff, b_t[T]);

		T += 1;
	}

	return res;
}


double GeneralizedSamplingDataNoDiscard::ComputeCCoeff(int S, int T)
{
	double res = 0.0;

//	printf("C(%d,%d)\n", S, T);

	int U = 0;
	while (U <= T)
	{
		//skip over U not in POWER(T)
		if ((U & ~T) != 0)
		{
			U += 1;
			continue;
		}

		//ip = T - U
		int ip = T & ~U;

		double coeff = 1.0;
		for (int i = 0; i < relations_no; i++)
		{
			// now treat bit in scan as in the big variance since these are variance terms
			if ((ip & (1 << i)) != 0)
				coeff *= -1.0;
		}

		int s_union_u = U | S;
		res += coeff * b_t[s_union_u];

//		printf("%d : %f %f\n", ip, coeff, b_t[s_union_u]);

		U += 1;
	}

	return res;
}



double GeneralizedSamplingDataNoDiscard::ComputeTCoeff()
{
	int T_no = threshold->size();

	double T = gsl_pow_int((*threshold)[T_no - 1], relations_no - constant_relations);
	for (int i = 0; i < relations_no; i++)
	{
		T *= (*bernoulli_p[i])[T_no - 1];
	}

	if (T_no == 1)
	{
		return T;
	}


	double t0, t1, t2;

	t1 = 1.0;
	for (int i = 0; i < relations_no; i++)
	{
		t1 *= (*bernoulli_p[i])[0];
	}

	for (int i = 0; i < T_no - 1; i++)
	{
		t0 = gsl_pow_int((*threshold)[i], relations_no - constant_relations);

		t2 = 1.0;
		for (int j = 0; j < relations_no; j++)
		{
			t2 *= (*bernoulli_p[j])[i + 1];
		}

		T += t0 * (t1 - t2);

		t1 = t2;
	}

	return T;
}


double GeneralizedSamplingDataNoDiscard::ComputeTVarCoeff(int tc)
{
	int T_no = threshold->size();
	double T = 1.0;

	//first, deal with the Bernoulli probabilities
	for (int j = 0; j < relations_no; j++)
	{
		//S contains j
		if ((tc & (1 << j)) != 0)
		{
			T *= (*bernoulli_p[j])[T_no - 1];
		}
		//S does not contain j
		else
		{
			T *= ((*bernoulli_p[j])[T_no - 1] * (*bernoulli_p[j])[T_no - 1]);
		}
	}

	//second, deal with the threashold probabilities
	for (int j = 0; j < relations_no; j++)
	{
		// skip over constant relations
		if ((constant_mask & (1 << j)) != 0)
			continue;

		if ((tc & (1 << j)) != 0)
		{
			T *= (*threshold)[T_no - 1];
		}
		else
		{
			T *= ((*threshold)[T_no - 1] * (*threshold)[T_no - 1]);
		}
	}

	if (T_no == 1)
	{
		return T;
	}


	double t0, t1, t2;

	t1 = 1.0;
	for (int j = 0; j < relations_no; j++)
	{
		//S contains j
		if ((tc & (1 << j)) != 0)
		{
			t1 *= (*bernoulli_p[j])[0];
		}
		//S does not contain j
		else
		{
			t1 *= ((*bernoulli_p[j])[0] * (*bernoulli_p[j])[0]);
		}
	}

	for (int i = 0; i < T_no - 1; i++)
	{
		t0 = 1.0;
		for (int j = 0; j < relations_no; j++)
		{
			// skip over constant relations
			if ((constant_mask & (1 << j)) != 0)
				continue;
	
			if ((tc & (1 << j)) != 0)
			{
				t0 *= (*threshold)[i];
			}
			else
			{
				t0 *= ((*threshold)[i] * (*threshold)[i]);
			}
		}

		t2 = 1.0;
		for (int j = 0; j < relations_no; j++)
		{
			//S contains j
			if ((tc & (1 << j)) != 0)
			{
				t2 *= (*bernoulli_p[j])[i + 1];
			}
			//S does not contain j
			else
			{
				t2 *= ((*bernoulli_p[j])[i + 1] * (*bernoulli_p[j])[i + 1]);
			}
		}

		T += t0 * (t1 - t2);

		t1 = t2;
	}

	return T;
}


void GeneralizedSamplingDataNoDiscard::Print()
{
	int n = threshold->size();

	cout << "Threshold:\n";
	for (int i = 0; i < n; i++)
	{
		printf("%10.8f\n", (*threshold)[i]);
	}

	cout << "Bernoulli probabilities:\n";
	for (int i = 0; i < relations_no; i++)
	{
		for (int j = 0; j < n; j++)
		{
			printf("%10.8f\t", (*bernoulli_p[i])[j]);
		}

		cout << "\n";
	}

	cout << "A : ";
	printf("%30.10f\n", a);

	cout << "B :\n";
	for (int i = 0; i < (1 << relations_no); i++)
	{
		printf("%30.10f\n", b_t[i]);
	}
}








SamplingEstimate::SamplingEstimate(int relations_no)
{
	this->relations_no = relations_no;

	coefficients = new (std::nothrow) double[1 << this->relations_no];
	if (coefficients == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	unbiased_coefficients = new (std::nothrow) double[1 << this->relations_no];
	if (unbiased_coefficients == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	unbiased_coefficients_cur = new (std::nothrow) double[1 << this->relations_no];
	if (unbiased_coefficients_cur == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	is_unbiased_computed = new (std::nothrow) int[1 << this->relations_no];
	if (is_unbiased_computed == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		coefficients[i] = 0.0;
		unbiased_coefficients[i] = 0.0;
		unbiased_coefficients_cur[i] = 0.0;

		is_unbiased_computed[i] = 0;
	}

	avg_no = 0;
}


SamplingEstimate::~SamplingEstimate()
{
	delete [] coefficients;
	coefficients = NULL;

	delete [] unbiased_coefficients;
	unbiased_coefficients = NULL;

	delete [] unbiased_coefficients_cur;
	unbiased_coefficients_cur = NULL;

	delete [] is_unbiased_computed;
	is_unbiased_computed = NULL;

	relations_no = 0;
	avg_no = 0;
}



void SamplingEstimate::GenerateStatistics(int tuples_no, int **tuples, double *func)
{
	TupleRepository *repository = new (std::nothrow) TupleRepository(relations_no, tuples_no, tuples, func);
	if (repository == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << relations_no); i++)
	{
		repository->Sort(i);
		coefficients[i] = repository->ComputeOrderCoef();
	}

	delete repository;
}



double SamplingEstimate::ComputeBiasedCoef(int order, SimpleSamplingData *data)
{
	double coeff = 1.0;
	for (int i = 0; i < relations_no; i++)
	{
		//bit i is set in order
		if ((order & (1 << i)) != 0)
			coeff *= (1.0 / data->e[i]);
		else
			coeff *= (1.0 / (data->e[i] * data->e[i]));
	}

	double res = coefficients[order] * coeff;
	return res;
}



double SamplingEstimate::ComputeUnbiasedCoef(int order, SimpleSamplingData *data)
{
	//check if the order coefficient is already computed
	if (is_unbiased_computed[order] != 0)
	{
		//cout << "Coefficient " << order << " was computed previously !!!\n";
		double res = unbiased_coefficients_cur[order];
		return res;
	}

	//first compute the unbiased coefficients for all the super-terms of the order term
	double unbiased_rest = 0.0;

	int scan = order;
	while (scan < (1 << relations_no))
	{
		for (scan = scan + 1; (scan < (1 << relations_no)) && ((order | scan) == order); scan++);
		scan = scan | order;
		if (scan >= (1 << relations_no))
			break;

		double agg_val = ComputeUnbiasedCoef(scan, data);
		
		//compute the coefficient in front c_{S,T)
		double coeff = 1.0;
		for (int i = 0; i < relations_no; i++)
		{
			//if bit is set in order than do nothing
			if ((order & (1 << i)) != 0)
				continue;
		
			// now treat bit in scan as in the big variance since these are variance terms
			if ((scan & (1 << i)) != 0)
				coeff *= data->b_e2[i];
			else
				coeff *= data->a_e2[i];
		}

		unbiased_rest += agg_val * coeff;
	}

	//compute the coefficient in front of the order term c_{\empty,S}
	double multiplier = 1.0;
	for (int i = 0; i < relations_no; i++)
	{
		if ((order & (1 << i)) == 0)
			multiplier *= data->a_e2[i];
	}
	
	double unbiased_est = ComputeBiasedCoef(order, data);
	unbiased_est -= unbiased_rest;
	unbiased_est /= multiplier;

	unbiased_coefficients_cur[order] = unbiased_est;
	is_unbiased_computed[order] = 1;

	return unbiased_est;
}



void SamplingEstimate::ComputeUnbiasedCoefs(SimpleSamplingData *data)
{
	for (int i = 0; i < (1 << relations_no); i++)
	{
		is_unbiased_computed[i] = 0;
	}

	for (int i = (1 << relations_no) - 1; i >= 0; i--)
	{
		unbiased_coefficients_cur[i] = ComputeUnbiasedCoef(i, data);
		is_unbiased_computed[i] = 1;
	}

	for (int i = 0; i < (1 << relations_no); i++)
	{
		double c_coef = unbiased_coefficients_cur[i];
		unbiased_coefficients[i] = (unbiased_coefficients[i] * avg_no + c_coef) / ((double)avg_no + 1.0);
//		unbiased_coefficients[i] = unbiased_coefficients_cur[i];
	}

	avg_no += 1;
}



double SamplingEstimate::ComputeExpectation(SimpleSamplingData *data, int tuples_no, double *func)
{
	double res = 0;
	
	for (int i = 0; i < tuples_no; i++)
	{
		res += func[i];
	}

	for (int i = 0; i < data->relations_no; i++)
	{
		res *= 1.0 / data->e[i];
	}
	
	return res;
}




double SamplingEstimate::ComputeVariance(SimpleSamplingData *data)
{
	//this is the square of the expected value
	double var = -unbiased_coefficients[0];
	
	for (int j = 0; j < (1 << relations_no); j++)
	{
		//compute the coefficient
		double coeff = 1.0;
		for (int i = 0; i < data->relations_no; i++)
		{
			if ((j & (1 << i)) != 0)
				coeff *= data->b_e2[i];
			else 
				coeff *= data->a_e2[i];
		}
		
		var += unbiased_coefficients[j] * coeff;
	}
	
	return var;
}


void SamplingEstimate::PrintCoefficients()
{
	cout << "Current unbiased coefficients:\n";
	for (int i = (1 << relations_no) - 1; i >= 0; i--) {
		cout << i << "\t" << unbiased_coefficients_cur[i] << endl;
	}

	cout << "Unbiased coefficients:\n";
	for (int i = (1 << relations_no) - 1; i >= 0; i--) {
		cout << i << "\t" << unbiased_coefficients[i] << endl;
	}
}





GeneralizedSamplingEstimate::GeneralizedSamplingEstimate(int relations_no, int tuples_stable)
{
	this->relations_no = relations_no;
	this->tuples_stable = tuples_stable;

	coefficients = new (std::nothrow) double[1 << this->relations_no];
	if (coefficients == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	unbiased_coefficients = new (std::nothrow) double[1 << this->relations_no];
	if (unbiased_coefficients == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	is_unbiased_computed = new (std::nothrow) int[1 << this->relations_no];
	if (is_unbiased_computed == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	v_term = new (std::nothrow) VarianceTerm*[1 << this->relations_no];
	if (v_term == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		coefficients[i] = 0.0;
		unbiased_coefficients[i] = 0.0;

		is_unbiased_computed[i] = 0;

		v_term[i] = new VarianceTerm(i, this->relations_no, VARIANCE_HASH_SIZE);
	}


	expectation_estimate = 0.0;

	tuples_in = 0;
	is_variance_stable = 0;
	deletes_no = 0;
}


GeneralizedSamplingEstimate::~GeneralizedSamplingEstimate()
{
	delete [] coefficients;
	coefficients = NULL;

	delete [] unbiased_coefficients;
	unbiased_coefficients = NULL;

	delete [] is_unbiased_computed;
	is_unbiased_computed = NULL;

	if (is_variance_stable == 0)
	{
		for (int i = 0; i < (1 << this->relations_no); i++)
		{
			delete v_term[i];
		}
		delete [] v_term;
	}

	relations_no = 0;
}



void GeneralizedSamplingEstimate::UpdateStatistics(int add_tuples_no, int **add_tuples, double *add_func, int del_tuples_no, int **del_tuples, double *del_func, GeneralizedSamplingData *data)
{
/*
	for (int j = 0; j < add_tuples_no; j++)
	{
		cout << "I : ";
		for (int i = 0; i < relations_no; i++)
		{
			cout << add_tuples[j][i] << "\t";
		}
		cout << "\n";
	}

	for (int j = 0; j < del_tuples_no; j++)
	{
		cout << "D : ";
		for (int i = 0; i < relations_no; i++)
		{
			cout << del_tuples[j][i] << "\t";
		}
		cout << "\n";
	}
*/


	//update the expectation part
	for (int j = 0; j < add_tuples_no; j++)
	{
		expectation_estimate += add_func[j];

//		cout << "Function value: " << add_func[j] << "\n";
	}

	for (int j = 0; j < del_tuples_no; j++)
	{
		expectation_estimate -= del_func[j];
	}



	//update the variance part
	if (is_variance_stable == 0)
	{
		for (int i = 0; i < (1 << relations_no); i++)
		{
			for (int j = 0; j < add_tuples_no; j++)
			{
				v_term[i]->AddTuple(add_tuples[j], relations_no, add_func[j]);
			}

			for (int j = 0; j < del_tuples_no; j++)
			{
				v_term[i]->RemoveTuple(del_tuples[j], relations_no, del_func[j]);
			}

			coefficients[i] = v_term[i]->GetEstimate();
		}

		ComputeUnbiasedCoefs(data);
	}


	if (del_tuples_no > 0)
		deletes_no += 1;

	tuples_in = tuples_in + add_tuples_no - del_tuples_no;
	if ((is_variance_stable == 0) && ((tuples_in >= tuples_stable) || (deletes_no > 0)))
	{
		for (int i = 0; i < (1 << this->relations_no); i++)
		{
			delete v_term[i];
			v_term[i] = NULL;
		}
		delete [] v_term;
		v_term = NULL;

		is_variance_stable = 1;

//		cout << "Variance is stable !!!\n";
	}
}



double GeneralizedSamplingEstimate::ComputeUnbiasedCoef(int S, GeneralizedSamplingData *data)
{
	//check if the S coefficient is already computed
	if (is_unbiased_computed[S] != 0)
	{
		//cout << "Coefficient " << S << " was computed previously !!!\n";
		double res = unbiased_coefficients[S];
		return res;
	}

	//first compute the unbiased coefficients for all the super-terms of the S term
	double unbiased_rest = 0.0;

	int scan = S;
	while (scan < (1 << relations_no))
	{
		//enumerates all supersets of S skipping the ones that are not
		for (scan = scan + 1; (scan < (1 << relations_no)) && ((S | scan) == S); scan++);
		scan = scan | S;
		if (scan >= (1 << relations_no))
			break;

		double agg_val = ComputeUnbiasedCoef(scan, data);

		//T = scan - S
		int T = scan & ~S;
		double coeff = data->ComputeCCoeff(S, T);

		unbiased_rest += agg_val * coeff;
	}

	//compute the coefficient in front of the S term c_{\empty,S}
	double c_S0 = data->ComputeCCoeff(S, 0);
//	double c_S0 = data->ComputeCCoeff(~S & complement_mask);
	
	double unbiased_est = coefficients[S];// / gsl_pow_int(data->a, 2);
	unbiased_est -= unbiased_rest;

//	cout << "Unbiased estimate(" << S << "): " << unbiased_est << "\n";

	unbiased_est /= c_S0;

//	cout << "Unbiased estimate(" << S << "): " << unbiased_est << "\n";

	unbiased_coefficients[S] = unbiased_est;
	is_unbiased_computed[S] = 1;

	return unbiased_est;
}



void GeneralizedSamplingEstimate::ComputeUnbiasedCoefs(GeneralizedSamplingData *data)
{
	if (is_variance_stable == 1)
	{
		cout << "Variance is already stable. ERROR !!!\n";
		exit(1);
	}


	for (int i = 0; i < (1 << relations_no); i++)
	{
		is_unbiased_computed[i] = 0;
	}

	for (int i = 0; i < (1 << relations_no); i++)
	{
		unbiased_coefficients[i] = ComputeUnbiasedCoef(i, data);
		is_unbiased_computed[i] = 1;
	}
}




double GeneralizedSamplingEstimate::ComputeExpectation(GeneralizedSamplingData *data)
{
	//if there are no tuples discovered yet, simply return 0
	if (tuples_in <= 0)
		return 0.0;


	double res = expectation_estimate / data->a;

	return res;
}




double GeneralizedSamplingEstimate::ComputeVariance(GeneralizedSamplingData *data)
{
	//if there are no tuples discovered yet, simply return 0
	if (tuples_in <= 0)
		return 0.0;


	//assume that unbised_coefficients contain most up to date estimates !!!!!!


	//this is the square of the expected value
	double var = -unbiased_coefficients[0];

	for (int j = 0; j < (1 << relations_no); j++)
	{
		//compute the coefficient
		double coeff = data->ComputeCCoeff(j) / gsl_pow_int(data->a, 2);
		
		var += unbiased_coefficients[j] * coeff;

//		cout << j << "\t" << data->threshold << "\t" << coeff  << "\t" << unbiased_coefficients[j] << "\t" << coefficients[j] << "\n";
	}

//	cout << "var=" << var << "\tstd=" << sqrt(var) << endl;

	return var;
}



double GeneralizedSamplingEstimate::ComputeCovariance(GeneralizedSamplingDataCov *data)
{
	//if there are no tuples discovered yet, simply return 0
	if (tuples_in <= 0)
		return 0.0;


	//assume that unbised_coefficients contain most up to date estimates !!!!!!


	//this is the E[X1]*E[X2] term
	double covar = -unbiased_coefficients[0];

	for (int j = 0; j < (1 << relations_no); j++)
	{
		//compute the coefficient
		double coeff = data->ComputeCCoeff(j) / (data->a1 * data->a2);
		
		covar += unbiased_coefficients[j] * coeff;

//		cout << j << "\t" << data->threshold << "\t" << coeff  << "\t" << unbiased_coefficients[j] << "\t" << coefficients[j] << "\n";
	}

//	cout << "var=" << var << "\tstd=" << sqrt(var) << endl;

	return covar;
}




void GeneralizedSamplingEstimate::FreezeCoefficients()
{
	//variance coefficients are not stable yet
	//freeze them anyway
	if (is_variance_stable == 0)
	{
		for (int i = 0; i < (1 << this->relations_no); i++)
		{
			delete v_term[i];
			v_term[i] = NULL;
		}
		delete [] v_term;
		v_term = NULL;

		is_variance_stable = 1;
	}
}



void GeneralizedSamplingEstimate::PrintCoefficients()
{
	for (int i = (1 << relations_no) - 1; i >= 0; i--) {
		cout << i << "\t" << coefficients[i] << endl;
	}
	
	for (int i = (1 << relations_no) - 1; i >= 0; i--) {
		cout << i << "\t" << unbiased_coefficients[i] << endl;
	}
}





GeneralizedSamplingEstimateNoDiscard::GeneralizedSamplingEstimateNoDiscard(int relations_no, int tuples_stable)
{
	this->relations_no = relations_no;
	this->tuples_stable = tuples_stable;

	coefficients = new (std::nothrow) double[1 << this->relations_no];
	if (coefficients == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	unbiased_coefficients = new (std::nothrow) double[1 << this->relations_no];
	if (unbiased_coefficients == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	is_unbiased_computed = new (std::nothrow) int[1 << this->relations_no];
	if (is_unbiased_computed == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	v_term = new (std::nothrow) VarianceTerm*[1 << this->relations_no];
	if (v_term == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < (1 << this->relations_no); i++)
	{
		coefficients[i] = 0.0;
		unbiased_coefficients[i] = 0.0;

		is_unbiased_computed[i] = 0;

		v_term[i] = new VarianceTerm(i, this->relations_no, VARIANCE_HASH_SIZE);
	}


	expectation_estimate = 0.0;

	tuples_in = 0;
	is_variance_stable = 0;
	deletes_no = 0;
}


GeneralizedSamplingEstimateNoDiscard::~GeneralizedSamplingEstimateNoDiscard()
{
	delete [] coefficients;
	coefficients = NULL;

	delete [] unbiased_coefficients;
	unbiased_coefficients = NULL;

	delete [] is_unbiased_computed;
	is_unbiased_computed = NULL;

	if (is_variance_stable == 0)
	{
		for (int i = 0; i < (1 << this->relations_no); i++)
		{
			delete v_term[i];
		}
		delete [] v_term;
	}

	relations_no = 0;
}



void GeneralizedSamplingEstimateNoDiscard::UpdateStatistics(int add_tuples_no, int **add_tuples, double *add_func, int del_tuples_no, int **del_tuples, double *del_func, GeneralizedSamplingDataNoDiscard *data)
{
	//update the expectation part
	for (int j = 0; j < add_tuples_no; j++)
	{
		expectation_estimate += add_func[j];

//		cout << "Function value: " << add_func[j] << "\n";
	}

	//update the variance part
	if (is_variance_stable == 0)
	{
		for (int i = 0; i < (1 << relations_no); i++)
		{
			for (int j = 0; j < add_tuples_no; j++)
			{
				v_term[i]->AddTuple(add_tuples[j], relations_no, add_func[j]);
			}

			coefficients[i] = v_term[i]->GetEstimate();
		}

		ComputeUnbiasedCoefs(data);
	}


	tuples_in = tuples_in + add_tuples_no;
	if ((is_variance_stable == 0) && (tuples_in >= tuples_stable))
	{
		for (int i = 0; i < (1 << this->relations_no); i++)
		{
			delete v_term[i];
			v_term[i] = NULL;
		}
		delete [] v_term;
		v_term = NULL;

		is_variance_stable = 1;

//		cout << "Variance is stable !!!\n";
	}
}



double GeneralizedSamplingEstimateNoDiscard::ComputeUnbiasedCoef(int S, GeneralizedSamplingDataNoDiscard *data)
{
	//check if the S coefficient is already computed
	if (is_unbiased_computed[S] != 0)
	{
		//cout << "Coefficient " << S << " was computed previously !!!\n";
		double res = unbiased_coefficients[S];
		return res;
	}

	//first compute the unbiased coefficients for all the super-terms of the S term
	double unbiased_rest = 0.0;

	int scan = S;
	while (scan < (1 << relations_no))
	{
		//enumerates all supersets of S skipping the ones that are not
		for (scan = scan + 1; (scan < (1 << relations_no)) && ((S | scan) == S); scan++);
		scan = scan | S;
		if (scan >= (1 << relations_no))
			break;

		double agg_val = ComputeUnbiasedCoef(scan, data);

		//T = scan - S
		int T = scan & ~S;
		double coeff = data->ComputeCCoeff(S, T);

		unbiased_rest += agg_val * coeff;
	}

	//compute the coefficient in front of the S term c_{\empty,S}
	double c_S0 = data->ComputeCCoeff(S, 0);
//	double c_S0 = data->ComputeCCoeff(~S & complement_mask);
	
	double unbiased_est = coefficients[S];// / gsl_pow_int(data->a, 2);
	unbiased_est -= unbiased_rest;

//	cout << "Unbiased estimate(" << S << "): " << unbiased_est << "\n";

	unbiased_est /= c_S0;

//	cout << "Unbiased estimate(" << S << "): " << unbiased_est << "\n";

	unbiased_coefficients[S] = unbiased_est;
	is_unbiased_computed[S] = 1;

	return unbiased_est;
}



void GeneralizedSamplingEstimateNoDiscard::ComputeUnbiasedCoefs(GeneralizedSamplingDataNoDiscard *data)
{
	if (is_variance_stable == 1)
	{
		cout << "Variance is already stable. ERROR !!!\n";
		exit(1);
	}


	for (int i = 0; i < (1 << relations_no); i++)
	{
		is_unbiased_computed[i] = 0;
	}

	for (int i = 0; i < (1 << relations_no); i++)
	{
		unbiased_coefficients[i] = ComputeUnbiasedCoef(i, data);
		is_unbiased_computed[i] = 1;
	}
}



double GeneralizedSamplingEstimateNoDiscard::ComputeExpectation(GeneralizedSamplingDataNoDiscard *data)
{
	//if there are no tuples discovered yet, simply return 0
	if (tuples_in <= 0)
		return 0.0;


	double res = expectation_estimate / data->a;

	return res;
}




double GeneralizedSamplingEstimateNoDiscard::ComputeVariance(GeneralizedSamplingDataNoDiscard *data)
{
	//if there are no tuples discovered yet, simply return 0
	if (tuples_in <= 0)
		return 0.0;


	//assume that unbised_coefficients contain most up to date estimates !!!!!!


	//this is the square of the expected value
	double var = -unbiased_coefficients[0];

	for (int j = 0; j < (1 << relations_no); j++)
	{
		//compute the coefficient
		double coeff = data->ComputeCCoeff(j) / gsl_pow_int(data->a, 2);
		
		var += unbiased_coefficients[j] * coeff;

//		cout << j << "\t" << data->threshold << "\t" << coeff  << "\t" << unbiased_coefficients[j] << "\t" << coefficients[j] << "\n";
	}

//	cout << "var=" << var << "\tstd=" << sqrt(var) << endl;

	return var;
}



void GeneralizedSamplingEstimateNoDiscard::FreezeCoefficients()
{
	//variance coefficients are not stable yet
	//freeze them anyway
	if (is_variance_stable == 0)
	{
		for (int i = 0; i < (1 << this->relations_no); i++)
		{
			delete v_term[i];
			v_term[i] = NULL;
		}
		delete [] v_term;
		v_term = NULL;

		is_variance_stable = 1;
	}
}



void GeneralizedSamplingEstimateNoDiscard::PrintCoefficients()
{
	for (int i = 0; i < (1 << relations_no); i++){
		cout << i << "\t" << coefficients[i] << endl;
	}
	
	for (int i = (1 << relations_no)-1; i>=0; i--){
		cout << i << "\t" << unbiased_coefficients[i] << endl;
	}
}






/* Problem: compute the probability:
   b_S=P[X_1<X_2<..<X_n & X_1'<X_2'<..<X_n' & \forall i\in S X_i=X_i']

   where X_i and X_i' are Uniform[0,1] random variables 
   (or any variables from a continous distribution)

   The key observation is the fact that all orderings of independent 
   random variables are equi-probable. The general strategy to determine 
   the probability is to count the number of compatible orders and to 
   divide this number by the number of possible orderings.

   The number of random variables is 2n-|S| since |S| random variables 
   are required to be equal.

   This function computes the term b_S. It computes the number 
   of good orders and then divides the number by the total possible orders

   The total number of orders is (2n-|S|)!
   Within a group of size k, the number of good orders is (2k)!/( k!)^2

   Groups are formed in the order 1,2,..,n
   A group starts the first time t_i != t_i' and ends when this no longer holds.
   t_i = t_i' when i\in S
   Essentialy k is the number of consecutive elements not in S.

   The code below determines the size of all the groups and maintains b_S while doing so. 
   Since the factorials involved can be rather large, ln(b_S) is computed and 
   exponentiated only at the end.

   This code depends on function: double gsl_sf_lnfact(unsigned int n) in the GSL package
*/

double Compute_Double_Prob(int S, int n, int constant_mask)
{
	double ln_bS = 0.0; // the ln(b_S) maintained
	
	int k = 0; // current group size
	int n_real = n;
	int S_size = 0;

	for (int i = 0; i < n; i++) {
		if ((constant_mask & (1 << i)) != 0) {
			n_real -= 1;
			continue;
		}

		if ((S & (1 << i)) != 0) {
			// the old group gets closed, it's contribution accounted for
			if (k > 0) { // if group has 0 size it contributes nothing
				ln_bS += gsl_sf_lnfact(2 * k) - 2.0 * gsl_sf_lnfact(k);
				// performs equivalent of *(2k)!/( k!)^2
				k = 0; // reset group
			}

			S_size += 1;
		}
		else {
			k++; // found group element. Do nothing
		}
	}
	
	// last group did not close; take care of that
	if (k > 0) { // if group has 0 size it contributes nothing
		ln_bS += gsl_sf_lnfact(2 * k) - 2.0 * gsl_sf_lnfact(k); // performs equivalent of *(2k)!/( k!)^2
	}
	
	// transform b_S into probability by dividing by (2n-|S|)!
	ln_bS -= gsl_sf_lnfact(2 * n_real - S_size);
	ln_bS = exp(ln_bS);

	return ln_bS;
}



double Compute_Double_Prob_Cov(int S, int n, int constant_mask, int n_chain, int how_many_in)
{
	double ln_bS = 0.0; // the ln(b_S) maintained
	
	int k = 0; // current group size
	int n_real = n_chain;
	int S_size = 0;

	int last_real = n - n_chain;
	int last_in_real = how_many_in;
	int last_out_real = n - n_chain - how_many_in;

	for (int i = 0; i < n_chain; i++) {
		if ((constant_mask & (1 << i)) != 0) {
			n_real -= 1;
			continue;
		}

		if ((S & (1 << i)) != 0) {
			// the old group gets closed, it's contribution accounted for
			if (k > 0) { // if group has 0 size it contributes nothing
				ln_bS += gsl_sf_lnfact(2 * k) - 2.0 * gsl_sf_lnfact(k);
				// performs equivalent of *(2k)!/( k!)^2
				k = 0; // reset group
			}

			S_size += 1;
		}
		else {
			k++; // found group element. Do nothing
		}
	}
	
	// last group did not close; take care of that
	if (k > 0) { // if group has 0 size it contributes nothing
		ln_bS += gsl_sf_lnfact(2 * k) - 2.0 * gsl_sf_lnfact(k); // performs equivalent of *(2k)!/( k!)^2
	}
	
	for (int i = n_chain; i < n_chain + how_many_in; i++) {
		if ((constant_mask & (1 << i)) != 0) {
			last_real -= 1;
			last_in_real -= 1;
		}
	}
	
	for (int i = n_chain + how_many_in; i < n; i++) {
		if ((constant_mask & (1 << i)) != 0) {
			last_real -= 1;
			last_out_real -= 1;
		}
	}

	ln_bS += gsl_sf_lnfact(last_real + last_in_real) - (gsl_sf_lnfact(last_real) + gsl_sf_lnfact(last_in_real)); // performs equivalent of *(2k)!/( k!)^2


	// transform b_S into probability by dividing by (2n-|S|)!
	ln_bS -= (gsl_sf_lnfact(2 * n_real + last_real + last_in_real - S_size) + gsl_sf_lnfact(last_out_real));
	ln_bS = exp(ln_bS);

//	printf("INC = %5d OUTC = %5d INI1 = %5d --> %5.3f\n", n_chain, n - n_chain, how_many_in, ln_bS);

	return ln_bS;
}


double Compute_Double_Prob_Cov_No_Const(int S, int n, int n_chain, int how_many_in)
{
	double ln_bS = 0.0; // the ln(b_S) maintained
	
	int k = 0; // current group size
	int S_size = 0;

	int last = n - n_chain;
	int last_in = how_many_in;
	int last_out = n - n_chain - how_many_in;

	for (int i = 0; i < n_chain; i++) {
		if ((S & (1 << i)) != 0) {
			// the old group gets closed, it's contribution accounted for
			if (k > 0) { // if group has 0 size it contributes nothing
				ln_bS += gsl_sf_lnfact(2 * k) - 2.0 * gsl_sf_lnfact(k);
				// performs equivalent of *(2k)!/( k!)^2
				k = 0; // reset group
			}

			S_size += 1;
		}
		else {
			k++; // found group element. Do nothing
		}
	}
	
	// last group did not close; take care of that
	if (k > 0) { // if group has 0 size it contributes nothing
		ln_bS += gsl_sf_lnfact(2 * k) - 2.0 * gsl_sf_lnfact(k); // performs equivalent of *(2k)!/( k!)^2
	}
	
	ln_bS += gsl_sf_lnfact(last + last_in) - (gsl_sf_lnfact(last) + gsl_sf_lnfact(last_in)); // performs equivalent of *(2k)!/( k!)^2


	// transform b_S into probability by dividing by (2n-|S|)!
	ln_bS -= (gsl_sf_lnfact(n + n_chain + last_in - S_size) + gsl_sf_lnfact(last_out));
	ln_bS = exp(ln_bS);

//	printf("INC = %5d OUTC = %5d INI1 = %5d --> %5.3f\n", n_chain, n - n_chain, how_many_in, ln_bS);

	return ln_bS;
}

