#ifndef _SAMPLING_H_
#define _SAMPLING_H_

//#define VARIANCE_TUPLES_STABLE 10000
#define VARIANCE_HASH_SIZE 10000
//#define VARIANCE_STABLE_COUNT 0


#include "StatStruct.h"

#include <vector>


using namespace std;


enum SamplingType {Bernoulli, WOR, WR};


//function computes the double probability that appears in the variance coefficients
//n is the total number of relations
//S is the subset for which the coefficient is computed
double Compute_Double_Prob(int S, int n, int constant_mask);
double Compute_Double_Prob_Cov(int S, int n, int constant_mask, int n_chain, int how_many_in);
double Compute_Double_Prob_Cov_No_Const(int S, int n, int n_chain, int how_many_in);



/* Abstract class to deal with sampling coefficients */
class SamplingData {
public:
	int relations_no;

	SamplingData();
	SamplingData(int relations_no);

	virtual ~SamplingData();

	/* compute c_S */
	virtual double ComputeCCoeff(int S) = 0;
	
	/* compute c_{S,T}; needed for unbiasing estimators */
	/* T has to be a strict superset of S; the method asumes that this holds without checking */
	virtual double ComputeCCoeff(int S, int T) = 0;
};



class SimpleSamplingData : public SamplingData {
public:
	SamplingType type;

	double *e;
	double *a_e2;
	double *b_e2;


	//generate the coefficients for Bernoulli sampling; p[i] is probability to select a tuple from relation i in the sample of relation i
	SimpleSamplingData(int relations_no, double *p);

	//generate the coefficients for wor sampling; rel_size[i] is the no of tuples in relation R[i]; samp_size[i] is the no of tuples in the sample of R[i]
	SimpleSamplingData(int relations_no, int *rel_size, int *samp_size);

	//copy constructor
	SimpleSamplingData(const SimpleSamplingData &sd);

	virtual ~SimpleSamplingData();


	virtual double ComputeCCoeff(int S);
	virtual double ComputeCCoeff(int S, int T);

	void Print();
};



class GeneralizedSamplingData : public SamplingData {
public:
	//scaling factor for expected value
	double a;

	//coefficients for the variance terms
	double *b_t;

	//the number of constant relations
	int constant_relations;
	//the mask of constant relations
	int constant_mask;

	double threshold;
	double *bernoulli_p;


	/* threshold is the cuttof used by TurboDBO, bernoulli_p are the Bernoulli probabilities for
	each relation and ConstantRelations is a bitmap indicating what relations are constant */
	//constant_mask specifies what relations are constant (a 1 in the bit vector stays for a constant relation
	GeneralizedSamplingData(int relations_no, int constant_relations, int constant_mask, double threshold, double *bernoulli_p);
	GeneralizedSamplingData(int relations_no, int constant_relations, int constant_mask, double threshold, vector<double> *bernoulli_p);
	
	virtual ~GeneralizedSamplingData();


	virtual double ComputeCCoeff(int S);
	virtual double ComputeCCoeff(int S, int T);

	void Print();
};



class GeneralizedSamplingDataCov : public SamplingData {
public:
	//scaling factor for expected value
	double a1, a2;

	//coefficients for the variance terms
	double *b_t;

	//the number of constant relations
	int constant_relations;
	//the mask of constant relations
	int constant_mask;

	double threshold1, threshold2;
	vector<double> *bernoulli_p1, *bernoulli_p2;


	/* threshold is the cuttof used by TurboDBO, bernoulli_p are the Bernoulli probabilities for
	each relation and ConstantRelations is a bitmap indicating what relations are constant */
	//constant_mask specifies what relations are constant (a 1 in the bit vector stays for a constant relation
	GeneralizedSamplingDataCov(int relations_no, int constant_relations, int constant_mask, double threshold1, double threshold2, vector<double> *bernoulli_p1, vector<double> *bernoulli_p2);
	
	virtual ~GeneralizedSamplingDataCov();


	virtual double ComputeCCoeff(int S);
	virtual double ComputeCCoeff(int S, int T);

	void Print();
};



class GeneralizedSamplingDataNoDiscard : public SamplingData {
public:
	//scaling factor for expected value
	double a;

	//coefficients for the variance terms
	double *b_t;

	//the number of constant relations
	int constant_relations;
	//the mask of constant relations
	int constant_mask;

	vector<double> *threshold;
	vector<double> **bernoulli_p;


	/* threshold is the cuttof used by TurboDBO, bernoulli_p are the Bernoulli probabilities for
	each relation and ConstantRelations is a bitmap indicating what relations are constant */
	//constant_mask specifies what relations are constant (a 1 in the bit vector stays for a constant relation
	GeneralizedSamplingDataNoDiscard(int relations_no, int constant_relations, int constant_mask, vector<double> *threshold, vector<double> **bernoulli_p);
	
	virtual ~GeneralizedSamplingDataNoDiscard();


	virtual double ComputeCCoeff(int S);
	virtual double ComputeCCoeff(int S, int T);

	void Print();

private:
	double ComputeTCoeff();
	double ComputeTVarCoeff(int tc);
};






class SamplingEstimate {
private:
	//compute the order biased coefficient
	double ComputeBiasedCoef(int order, SimpleSamplingData *data);

	//compute the order unbiased coefficient
	double ComputeUnbiasedCoef(int order, SimpleSamplingData *data);


public:
	int relations_no;
	
	double *coefficients;
	double *unbiased_coefficients;
	double *unbiased_coefficients_cur;

	int *is_unbiased_computed;

	int avg_no;


	SamplingEstimate(int relations_no);
	virtual ~SamplingEstimate();


	//generate the terms used for variance computation; repository is created
	void GenerateStatistics(int tuples_no, int **tuples, double *func);

	//compute all the unbiased coefficients
	void ComputeUnbiasedCoefs(SimpleSamplingData *data);

	//compute the expected value estimator
	double ComputeExpectation(SimpleSamplingData *data, int tuples_no, double *func);

	//compute the estimator for the variance
	double ComputeVariance(SimpleSamplingData *data);

	void PrintCoefficients();
};




class GeneralizedSamplingEstimate {
private:
	//compute the order unbiased coefficient
	double ComputeUnbiasedCoef(int order, GeneralizedSamplingData *data);

	//compute all the unbiased coefficients
	void ComputeUnbiasedCoefs(GeneralizedSamplingData *data);

public:
	int relations_no;
	
	double *coefficients;
	double *unbiased_coefficients;

	int *is_unbiased_computed;

	double expectation_estimate;

	int tuples_in;
	int is_variance_stable;
	int deletes_no;

	VarianceTerm **v_term;

	int tuples_stable;


	GeneralizedSamplingEstimate(int relations_no, int tuples_stable);
	virtual ~GeneralizedSamplingEstimate();


	//update the statistics (expectation and, maybe, variance)
	void UpdateStatistics(int add_tuples_no, int **add_tuples, double *add_func, int del_tuples_no, int **del_tuples, double *del_func, GeneralizedSamplingData *data);


	//compute the expected value estimator
	double ComputeExpectation(GeneralizedSamplingData *data);

	//compute the estimator for the variance
	double ComputeVariance(GeneralizedSamplingData *data);

	//compute the estimator for the covariance
	double ComputeCovariance(GeneralizedSamplingDataCov *data);

	//freeze coefficients since they are assumed to be stable
	//it is needed for grouping, when one level is done and the coefficients are not stable
	void FreezeCoefficients();

	void PrintCoefficients();
};




class GeneralizedSamplingEstimateNoDiscard {
private:
	//compute the order unbiased coefficient
	double ComputeUnbiasedCoef(int order, GeneralizedSamplingDataNoDiscard *data);

	//compute all the unbiased coefficients
	void ComputeUnbiasedCoefs(GeneralizedSamplingDataNoDiscard *data);


public:
	int relations_no;
	
	double *coefficients;
	double *unbiased_coefficients;

	int *is_unbiased_computed;

	double expectation_estimate;

	int tuples_in;
	int is_variance_stable;
	int deletes_no;

	VarianceTerm **v_term;

	int tuples_stable;


	GeneralizedSamplingEstimateNoDiscard(int relations_no, int tuples_stable);
	virtual ~GeneralizedSamplingEstimateNoDiscard();


	//update the statistics (expectation and, maybe, variance)
	void UpdateStatistics(int add_tuples_no, int **add_tuples, double *add_func, int del_tuples_no, int **del_tuples, double *del_func, GeneralizedSamplingDataNoDiscard *data);


	//compute the expected value estimator
	double ComputeExpectation(GeneralizedSamplingDataNoDiscard *data);

	//compute the estimator for the variance
	double ComputeVariance(GeneralizedSamplingDataNoDiscard *data);

	//freeze coefficients since they are assumed to be stable
	//it is needed for grouping, when one level is done and the coefficients are not stable
	void FreezeCoefficients();

	void PrintCoefficients();
};



#endif


