#include "StatStruct.h"

#include <iostream>
#include <string.h>
#include <stdlib.h>

using namespace std;



TupleRepository::TupleRepository(int relations_no, int tuples_no, int **tuples, double *func)
{
	this->relations_no = relations_no;
	this->tuples_no = tuples_no;

	this->func = new (std::nothrow) double[this->tuples_no];
	if (this->func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	memcpy(this->func, func, this->tuples_no * sizeof(double));

	//position 0 is the sorting order; position 1 is the number of relations; positions 2 to relations_no + 1 are the attribute identifiers for each of the relations; position relations_no + 2 is the initial index, necessary for summing the function values; this representation is required for sorting with qsort
	this->tuples = new (std::nothrow) int[this->tuples_no * (this->relations_no + 3)];
	if (this->tuples == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < this->tuples_no; i++)
	{
		this->tuples[i * (this->relations_no + 3)] = 0;
		this->tuples[i * (this->relations_no + 3) + 1] = this->relations_no;

		memcpy(this->tuples + i * (this->relations_no + 3) + 2, tuples[i], this->relations_no * sizeof(int));

		this->tuples[i * (this->relations_no + 3) + this->relations_no + 2] = i;
	}
}


TupleRepository::~TupleRepository()
{
	delete [] tuples;
	tuples = NULL;

	delete [] func;
	func = NULL;

	relations_no = 0;
	tuples_no = 0;
}



void TupleRepository::Sort(int order)
{
	for (int i = 0; i < tuples_no; i++)
	{
		tuples[i * (relations_no + 3)] = order;
	}

	qsort(tuples, tuples_no, sizeof(int) * (relations_no + 3), Compare_Tuples);
}


double TupleRepository::ComputeOrderCoef()
{
	double res = 0.0;
	double p_res = func[tuples[relations_no + 2]];

	int c_p = 0;
	for (int i = 1; i < tuples_no; i++)
	{
		if (Compare_Tuples(tuples + c_p * (relations_no + 3), tuples + i * (relations_no + 3)) == 0)
		{
			p_res += func[tuples[i * (relations_no + 3) + relations_no + 2]];
		}
		else
		{
			res += (p_res * p_res);

			p_res = func[tuples[i * (relations_no + 3) + relations_no + 2]];
			c_p = i;
		}
	}

	res += (p_res * p_res);

	return res;
}





VarianceFactor::VarianceFactor(int mask, int *v, int n)
{
	int i;
	int k = 0;

	for (i = 0; i < n; i++)
	{
		if ((mask & (1 << i)) != 0)
			k += 1;
	}

	this->n = k;

	this->v = new (std::nothrow) int[this->n];
	if (this->v == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	k = 0;
	for (i = 0; i < n; i++)
	{
		if ((mask & (1 << i)) != 0)
		{
			this->v[k] = v[i];
			k += 1;
		}
	}

	this->func = 0.0;
}


VarianceFactor::~VarianceFactor()
{
	delete [] v;
}


int VarianceFactor::Compare(VarianceFactor *f)
{
	if (this->n != f->n)
		return 1;

	int c_res = memcmp(this->v, f->v, sizeof(int) * this->n);

	return c_res;
}




VarianceTerm::VarianceTerm(int mask, int n, int repository_size)
{
	int i;
	int k = 0;

	this->mask = mask;

	for (i = 0; i < n; i++)
	{
		if ((mask & (1 << i)) != 0)
			k += 1;
	}

	this->n = k;

	this->repository_size = repository_size;
	this->repository = new (std::nothrow) vector<VarianceFactor*>[this->repository_size];
	if (this->repository == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->hash_func = new (std::nothrow) CWHash(this->n + 1, this->repository_size);
	if (this->hash_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->estimate = 0.0;
}


VarianceTerm::~VarianceTerm()
{
	delete hash_func;

	for (int i = 0; i < repository_size; i++)
	{
		for (int j = 0; j < repository[i].size(); j++)
		{
			VarianceFactor *f = (repository[i])[j];
			delete f;
			(repository[i])[j] = NULL;
		}

		repository[i].clear();
	}
	delete [] repository;
}


double VarianceTerm::GetEstimate()
{
	double res = estimate;
	if (this->n == 0)
	{
		res *= estimate;
	}

	return res;
}


void VarianceTerm::AddTuple(int *tuple, int n, double func) {
	//extreme case for all terms different; only sum squares
	if (this->n == 0)
	{
		estimate += func;
		return;
	}

	/* extract part of tuple projected on S*/
	VarianceFactor *new_fact = new (std::nothrow) VarianceFactor(mask, tuple, n);
	if (new_fact == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int bucket = hash_func->Compute_Hash(new_fact->v, new_fact->n);

	//search sequentially in the bucket for this factor
	int exists = 0;
	for (int i = 0; i < repository[bucket].size(); i++)
	{
		if (new_fact->Compare((repository[bucket])[i]) == 0)
		{
			//the factor already exists; just update the estimate
			VarianceFactor *c_factor = (repository[bucket])[i];

			double cf = c_factor->func;
			estimate -= (cf * cf);
			cf += func;
			estimate += (cf * cf);

			c_factor->func += func;

			delete new_fact;

			exists = 1;
			break;
		}
	}

	if (exists == 0)
	{
		//the factor does not exist, so append to the repository
		new_fact->func = func;
		repository[bucket].push_back(new_fact);

		estimate += (func * func);
	}
}


void VarianceTerm::RemoveTuple(int *tuple, int n, double func) {
	//extreme case for all terms different; only sum squares
	if (this->n == 0)
	{
		estimate -= func;
		return;
	}

	/* extract part of tuple projected on S*/
	VarianceFactor *new_fact = new (std::nothrow) VarianceFactor(mask, tuple, n);
	if (new_fact == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int bucket = hash_func->Compute_Hash(new_fact->v, new_fact->n);

	//search sequentially in the bucket for this factor
	int exists = 0;
	for (int i = 0; i < repository[bucket].size(); i++)
	{
		if (new_fact->Compare((repository[bucket])[i]) == 0)
		{
			//the factor already exists; just update the estimate
			VarianceFactor *c_factor = (repository[bucket])[i];

			double cf = c_factor->func;
			estimate -= (cf * cf);
			cf -= func;
			estimate += (cf * cf);

			c_factor->func -= func;

			exists = 1;
			break;
		}
	}

	if (exists == 0)
	{
		//the factor does not exist, ERROR
		cout << "A tuple that does not exist is required to be deleted in STATISTICS!!!\n";

		cout << "Tuple that does not exist:\n";
		for (int i = 0; i < n; i++)
		{
			cout << tuple[i] << "\t";
		}
		cout << "\n";
	}

	delete new_fact;
}





//compare two tuples based on the order specified inside them
int Compare_Tuples(const void *r1, const void *r2)
{
	int *a, *b;

	a = (int*)r1;
	b = (int*)r2;

	int order = *a;
	int rels_no = *(a + 1);

	int res = 0;

	for (int i = 0; i < rels_no; i++)
	{
		if ((order & (1 << i)) != 0)
		{
			//attribute i is included in sorting
			if (a[i + 2] > b[i + 2])
			{
				res = 1;
				break;
			}
			else if (a[i + 2] < b[i + 2])
			{
				res = -1;
				break;
			}
		}
	}

	return res;
}
