#include "Statistics.h"
#include "StatisticsLevel.h"
#include "SamplingEstimate.h"

#include <vector>
#include <iostream>
#include <gsl/gsl_cdf.h>
#include <stdlib.h>
#include <math.h>



Statistics::Statistics(int tuples_stable, int var_estimates_needed)
{
	levels_no = 0;
	is_last_level = 0;

	this->tuples_stable = tuples_stable;
	this->var_estimates_needed = var_estimates_needed;

	e = new (std::nothrow) vector<double>;
	if (e == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	var = new (std::nothrow) vector<double>;
	if (var == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	cur_level = NULL;
}


Statistics::~Statistics()
{
	e->clear();
	delete e;
	e = NULL;

	var->clear();
	delete var;
	var = NULL;

	delete cur_level;
	cur_level = NULL;

	levels_no = 0;
	is_last_level = 0;
}



void Statistics::AddLevel(int numRels, int isLastLevel)
{
	delete cur_level;

	if (isLastLevel == 1)
	{
		cur_level = new StatisticsUpLevel(numRels);
	}
	else
	{
		cur_level = new StatisticsLowLevel(numRels, var_estimates_needed);
	}

	e->push_back(0.0);
	var->push_back(0.0);

	levels_no += 1;
	is_last_level = isLastLevel;
}


void Statistics::AddGeneralLevel(int numRels, int isLastLevel)
{
	delete cur_level;


	if (isLastLevel == 1)
	{
		cur_level = new StatisticsUpLevel(numRels);
	}
	else
	{
		//cur_level = new StatisticsGeneralLevel(numRels, tuples_stable);
		cur_level = new StatisticsGeneralLevelCov(numRels, tuples_stable);
		//cur_level = new StatisticsGeneralLevelNoDiscard(numRels, tuples_stable);
	}


//	cur_level = new StatisticsGeneralLevel(numRels);

	e->push_back(0.0);
	var->push_back(0.0);

	levels_no += 1;
	is_last_level = isLastLevel;
}



int Statistics::GetEst(double &estimate, double &variance, double &low, double &high, double confLevel)
{
	(*e)[levels_no - 1] = cur_level->GetExpectation();
	(*var)[levels_no - 1] = cur_level->GetVariance();


/*
	cout << "Level estimates:\n";
	for (int i = 0; i < levels_no; i++)
	{
		cout << i << " : " << (*e)[i] << " " << (*var)[i] << "\n";
	}
	cout << "\n";
*/


	int ret = Solve_Optimization_Levels(e, var, levels_no, estimate, variance);
	if (ret == 0) // || ((*var)[levels_no - 1] < variance))
	{
		estimate = (*e)[levels_no - 1];
		variance = (*var)[levels_no - 1];
	}


//	estimate = (*e)[levels_no - 1];
//	variance = (*var)[levels_no - 1];


	//compute the confidence intervals based on the CLT
	double low_sigma = (1.0 - confLevel) / 2.0;
	double high_sigma = confLevel + low_sigma;

	low = estimate + gsl_cdf_gaussian_Pinv(low_sigma, sqrt(variance));
	high = estimate + gsl_cdf_gaussian_Pinv(high_sigma, sqrt(variance));

	return 0;
}


int Statistics::GetEst(double &estimate, double &variance, double &low, double &high, double confLevel, double p)
{
	(*e)[levels_no - 1] = cur_level->GetExpectation(p);
	(*var)[levels_no - 1] = cur_level->GetVariance(p);

/*
	cout << "Level estimates:\n";
	for (int i = 0; i < levels_no; i++)
	{
		cout << i << " : " << (*e)[i] << " " << (*var)[i] << "\n";
	}
	cout << "\n";
*/


	int ret = Solve_Optimization_Levels(e, var, levels_no, estimate, variance);
	if (ret == 0) // || ((*var)[levels_no - 1] < variance))
	{
		estimate = (*e)[levels_no - 1];
		variance = (*var)[levels_no - 1];
	}


//	estimate = (*e)[levels_no - 1];
//	variance = (*var)[levels_no - 1];

	//compute the confidence intervals based on the CLT
	double low_sigma = (1.0 - confLevel) / 2.0;
	double high_sigma = confLevel + low_sigma;

	low = estimate + gsl_cdf_gaussian_Pinv(low_sigma, sqrt(variance));
	high = estimate + gsl_cdf_gaussian_Pinv(high_sigma, sqrt(variance));

	return 0;
}


double Statistics::WhatIfVariance(double *proposedSamplingRatios)
{
	double res = cur_level->WhatIfVariance(proposedSamplingRatios);
	return res;
}



int Statistics::ProcessTuples(int **tuples, double *funcVals, int numTuples, SimpleSamplingData *data)
{
	int res = cur_level->ProcessTuples(tuples, funcVals, numTuples, data);
	return res;
}


int Statistics::ProcessTuples(double *funcVals, int numTuples, SimpleSamplingData *data)
{
	int res = cur_level->ProcessTuples(funcVals, numTuples, data);
	return res;
}


void Statistics::ProcessTuples(SearchResults *newResults, double p)
{
	cur_level->ProcessTuples(newResults, p);
}


void Statistics::FreezeCoefficients()
{
	cur_level->FreezeCoefficients();
}


void Statistics::Print()
{
	for (int i = 0; i < levels_no; i++)
	{
		cout << "E[" << i << "] = " << (*e)[i] << "\n";
		cout << "VAR[" << i << "] = " << (*var)[i] << "\n";
	}
}


void Statistics::Print_Level()
{
	cout << "Level:\n";
	cur_level->Print();
}




int Solve_Optimization_Levels(vector<double> *e, vector<double> *var, int n, double &opt_e, double &opt_var)
{
	opt_e = 0.0;
	opt_var = 0.0;

	for (int i = 0; i < n; i++)
	{
		if ((*e)[i] != 0.0)
		{
			if (fabs((*var)[i] / (*e)[i]) < 1.0e-5)
			{
				opt_e = (*e)[i];
				opt_var = (*var)[i];

				return 1;
			}
		}
		if ((*var)[i] > 0.0)
		{
			opt_e += (*e)[i] / (*var)[i];
			opt_var += 1.0 / (*var)[i];
		}
	}

	if (opt_var != 0.0)
	{
		opt_e = opt_e / opt_var;
		opt_var = 1.0 / opt_var;

		return 1;
	}

	return 0;
}
