#ifndef _STATISTICS
#define _STATISTICS

#include "SamplingEstimate.h"
#include "StatisticsLevel.h"
#include "DEFS.h"

#include <vector>


using namespace std;


int Solve_Optimization_Levels(vector<double> *e, vector<double> *var, int n, double &opt_e, double &opt_var);


class Statistics {
public:
	// the constructor.  The param to the constuctor is the maximum
	// normalized standard deviation that is tolerated in the variance
	// estimate.  For example, if maxErr is 0.1, then this means that
	// the ratio of the variance estiamte to the standard deviation of
	// the variance estimate needs to be less than 10%.  We obtain one
	// estimate for the variance every time the first version of TakeTuples
	// is called; see below.  The normalized standard deviation of the
	// variance estimate can then easily be computed by bootstrap sampling
	// of all of the individual variance estimates.
	Statistics(int tuples_stable, int var_estimates_needed);
	virtual ~Statistics();


	// this lets the statistics object know that it is about to begin
	// a new levelwise step.  The parameter tells the object how many
	// relations are going to be used as input into this levelwise step
	// Note that if numRels is 1, then the object knows that this is the
	// last (topmost) levelwise step, and it performs its variance estiamtion
	// using the procedure outlined in the "Estimation at the Topmost Level"
	// section of the paper
	void AddLevel(int numRels, int isLastLevel);
	void AddGeneralLevel(int numRels, int isLastLevel);

	// this returns the current estimate, along with the low and high
	// confidence bounds at the given confidence level.  To get this
	// esimimate, all of the estimates from all of the various levelwise
	// steps are combined using a weighted average, with the weights 
	// computed as described in the paper
	int GetEst(double &estimate, double &variance, double &low, double &high, double confLevel);
	int GetEst(double &estimate, double &variance, double &low, double &high, double confLevel, double p);

	//needed in the IMJ for determining the dropping
	double WhatIfVariance(double *proposedSamplingRatios);

	// this takes another set of tuples into the statistics object,
	// and uses them to update *both* the estimate and the variance
	// estimate.  This set of tuples will be the set returned by the
	// in-memory join, after the collapsing operation has been performed.
	// The first parameter gives a list of pointers to the tuples; each
	// tuple t is an array of integers that tells us the tuple ID of the
	// tuple from each input relation that was used to form t.  The second
	// parameter tells us the number of tuples in this list.  The third
	// parameter gives the aggregate function values associated with each
	// tuple (that is, funcVals[i] is associated with tuples[i]).  Finally,
	// the last parameter tells us the fraction of each input relation that
	// was used to produce these results.  If probs[2] = 0.45, this means
	// that 45% of the third input relation was joined to produce tuples
	// and funcVals 
	// the return value is a zero if the statistcs module still does not
	// think that it has been given enough data to make an accurate
	// estimate of the variance (accuracy) of the estimate.  In this case,
	// more full tuples should be fed into the statistics module so that
	// it can update the variance estimate and make it more accurate.
	// Otherwise, if the variance estimate is accurate enough, a one is
	// returned.
	int ProcessTuples(int **tuples, double *funcVals, int numTuples, SimpleSamplingData *data);
		
	// this version works for the top-most level, where no tuples need to be passed
	int ProcessTuples(double *funcVals, int numTuples, SimpleSamplingData *data);

	//p is the sweeping line
	void ProcessTuples(SearchResults *newResults, double p);

	void FreezeCoefficients();

	void Print();
	void Print_Level();


	vector<double> *e;
	vector<double> *var;

	int levels_no;
	int is_last_level;

	int tuples_stable;
	int var_estimates_needed;

	StatisticsLevel *cur_level;
};


#endif

